# Copyright(C) 2011-2021 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from base64 import b64decode, b64encode
import binascii

from woob.capabilities.paste import CapPaste


class BasePasteModule(CapPaste):
    EXPIRATIONS = {}
    """
    List of expirations and their corresponding remote codes (any type can be used).
    The expirations, i.e. the keys, are integers representing the duration
    in seconds. There also can be one False key, for the "forever" expiration.
    """

    def get_closest_expiration(self, max_age):
        """
        Get the expiration closest (and less or equal to) max_age (int, in seconds).
        max_age set to False means we want it to never expire.

        @return int or False if found, else None
        """
        # "forever"
        if max_age is False and False in self.EXPIRATIONS:
            return max_age
        # get timed expirations, longest first
        expirations = sorted([e for e in self.EXPIRATIONS if e is not False], reverse=True)
        # find the first expiration that is below or equal to the maximum wanted age
        for e in expirations:
            if max_age is False or max_age >= e:
                return e


def image_mime(data_base64, supported_formats=('gif', 'jpeg', 'png')):
    """
    Return the MIME type of an image or None.

    :param data_base64: data to detect, base64 encoded
    :type data_base64: str
    :param supported_formats: restrict list of formats to test
    """
    try:
        beginning = b64decode(data_base64[:24])
    except binascii.Error:
        return None

    if 'gif' in supported_formats and b'GIF8' in beginning:
        return 'image/gif'
    elif 'jpeg' in supported_formats and b'JFIF' in beginning:
        return 'image/jpeg'
    elif 'png' in supported_formats and b'\x89PNG' in beginning:
        return 'image/png'
    elif 'xcf' in supported_formats and b'gimp xcf' in beginning:
        return 'image/x-xcf'
    elif 'pdf' in supported_formats and b'%PDF' in beginning:
        return 'application/pdf'
    elif 'tiff' in supported_formats and (b'II\x00\x2a' in beginning or
          b'MM\x2a\x00' in beginning):
        return 'image/tiff'


def bin_to_b64(b):
    return b64encode(b).decode('ascii')



# Copyright(C) 2010-2013 Bezleputh
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from .base import BaseObject, Capability, StringField
from .date import DateField

__all__ = ['BaseJobAdvert', 'CapJob']


class BaseJobAdvert(BaseObject):
    """
    Represents a job announce.
    """
    publication_date = DateField('Date when the announce has been published')
    society_name =     StringField('Name of the society taht published the announce')
    place =            StringField('Place where the job take place')
    job_name =         StringField('Name of the job')
    title =            StringField('Title of the announce')
    contract_type =    StringField('Type of the contrat : CDI, CDD')
    pay =              StringField('Amount of the salary')
    description =      StringField('Description of the job')
    formation =        StringField('Required formation')
    experience =       StringField('Required experience')

    def __str__(self):
        message = '\r\n-- Advert --\r\n'
        message += 'id : %s\r\n' % self.id
        message += 'url : %s\r\n' % self.url
        message += 'publication_date : %s\r\n' % self.publication_date
        message += 'society_name : %s\r\n' % self.society_name
        message += 'place : %s\r\n' % self.place
        message += 'job_name : %s\r\n' % self.job_name
        message += 'title : %s\r\n' % self.title
        message += 'contract_type : %s\r\n' % self.contract_type
        message += 'pay : %s\r\n' % self.pay
        message += 'description : %s\r\n' % self.description
        message += 'formation : %s\r\n' % self.formation
        message += 'experience : %s\r\n' % self.experience
        return message

    @classmethod
    def id2url(cls, _id):
        """Overloaded in child classes provided by backends."""
        raise NotImplementedError()

    @property
    def page_url(self):
        """
        Get page URL of the announce.
        """
        return self.id2url(self.id)


class CapJob(Capability):
    """
    Capability of job annouce websites.
    """

    def search_job(self, pattern=None):
        """
        Iter results of a search on a pattern.

        :param pattern: pattern to search on
        :type pattern: str
        :rtype: iter[:class:`BaseJobAdvert`]
        """
        raise NotImplementedError()

    def advanced_search_job(self):
        """
         Iter results of an advanced search

        :rtype: iter[:class:`BaseJobAdvert`]
        """

    def get_job_advert(self, _id, advert=None):
        """
        Get an announce from an ID.

        :param _id: id of the advert
        :type _id: str
        :param advert: the advert
        :type advert: BaseJobAdvert
        :rtype: :class:`BaseJobAdvert` or None if not found.
        """
        raise NotImplementedError()



# Copyright(C) 2023      Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.
"""
This module provides tools to bypass cloudflare websites with your browser.

To use it, add :class:`CloudScraperMixin` as a mixin of your browser class.
"""

try:
    from cloudscraper import CloudScraper
except ImportError as exc:
    raise ImportError('Please install cloudscraper') from exc


__all__ = ['CloudScraperSession', 'CloudScraperMixin']


class CloudScraperSession(CloudScraper):
    def send(self, *args, **kwargs):
        callback = kwargs.pop('callback', lambda future, response: response)
        is_async = kwargs.pop('is_async', False)

        if is_async:
            raise ValueError('Async requests are not supported')

        resp = super().send(*args, **kwargs)

        return callback(self, resp)


class CloudScraperMixin:
    def _create_session(self):
        return CloudScraperSession()

    def _setup_session(self, profile):
        session = self._create_session()

        session.hooks['response'].append(self.set_normalized_url)
        if self.responses_dirname is not None:
            session.hooks['response'].append(self.save_response)

        self.session = session



# Copyright(C) 2020  Budget Insight
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from collections import OrderedDict


__all__ = ['sorted_documents', 'merge_iterators']


def sorted_documents(iterable):
    """Sort an iterable of documents in reverse chronological order"""
    return sorted(iterable, reverse=True, key=lambda doc: doc.date)


def merge_iterators(*iterables):
    """Merge documents iterators keeping sort order.

    Each iterator must already be sorted in reverse chronological order.
    """

    def keyfunc(kv):
        return kv[1].date

    its = OrderedDict((iter(it), None) for it in iterables)
    for k in list(its):
        try:
            its[k] = next(k)
        except StopIteration:
            del its[k]

    while its:
        k, v = max(its.items(), key=keyfunc)
        yield v

        try:
            its[k] = next(k)
        except StopIteration:
            del its[k]



# Copyright(C) 2010-2011 Christophe Benz, Romain Bignon, John Obbele, Nicolas Duhamel
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from io import BytesIO
import os
from shutil import which
import subprocess
from urllib.parse import urlparse

import requests

from woob.capabilities.video import CapVideo, BaseVideo
from woob.capabilities.base import empty
from woob.tools.application.repl import ReplApplication, defaultcount
from woob.tools.application.media_player import InvalidMediaPlayer, MediaPlayer, MediaPlayerNotFound
from woob.tools.application.formatters.iformatter import PrettyFormatter

from .image2xterm import image2xterm, get_term_size

__all__ = ['AppVideo']


class VideoListFormatter(PrettyFormatter):
    MANDATORY_FIELDS = ('id', 'title', 'duration', 'date')
    DISPLAYED_FIELDS = ('author', 'rating', 'thumbnail', 'rating_max')

    def get_title(self, obj):
        return obj.title

    def get_description(self, obj):
        if empty(obj.duration) and empty(obj.date):
            return None

        result = '%s' % (obj.duration or obj.date)
        if hasattr(obj, 'author') and not empty(obj.author):
            result += ' - %s' % obj.author
        if hasattr(obj, 'rating') and not empty(obj.rating):
            result += ' (%s/%s)' % (obj.rating, obj.rating_max)
        if hasattr(obj, 'thumbnail') and not empty(obj.thumbnail) and not empty(obj.thumbnail.data):
            result += '\n'
            result += image2xterm(BytesIO(obj.thumbnail.data), newsize=get_term_size())

        return result


class AppVideo(ReplApplication):
    APPNAME = 'video'
    VERSION = '3.7'
    COPYRIGHT = 'Copyright(C) 2010-YEAR Christophe Benz, Romain Bignon, John Obbele'
    DESCRIPTION = "Console application allowing to search for videos on various websites, " \
                  "play and download them and get information."
    SHORT_DESCRIPTION = "search and play videos"
    CAPS = CapVideo
    EXTRA_FORMATTERS = {'video_list': VideoListFormatter}
    COMMANDS_FORMATTERS = {'search': 'video_list',
                           'ls': 'video_list',
                           'playlist': 'video_list'}
    COLLECTION_OBJECTS = (BaseVideo, )
    PLAYLIST = []
    nsfw = True

    def __init__(self, *args, **kwargs):
        super(AppVideo, self).__init__(*args, **kwargs)
        self.player = MediaPlayer(self.logger)

    def main(self, argv):
        self.load_config()
        return super().main(argv)

    def download(self, video, dest, default=None):
        if not video.url:
            print('Error: the direct URL is not available.', file=self.stderr)
            return 4

        def check_exec(executable):
            if which(executable) is None:
                print('Please install "%s"' % executable, file=self.stderr)
                return False
            return True

        dest = self.obj_to_filename(video, dest, default)

        if video.url.startswith('rtmp'):
            if not check_exec('rtmpdump'):
                return 1
            args = ('rtmpdump', '-e', '-r', video.url, '-o', dest)
        elif video.url.startswith('mms'):
            if not check_exec('mimms'):
                return 1
            args = ('mimms', '-r', video.url, dest)
        elif 'm3u8' == video.ext:
            _dest, _ = os.path.splitext(dest)
            dest = '%s.%s' % (_dest, 'mp4')
            content = tuple()
            parsed_uri = urlparse(video.url)
            baseurl = '{uri.scheme}://{uri.netloc}'.format(uri=parsed_uri)
            for line in self.read_url(video.url):
                line = line.decode('utf-8')
                if not line.startswith('#'):
                    if not line.startswith('http'):
                        line = '%s%s' % (baseurl, line)
                    content += (line,)

            args = ('wget', '-nv',) + content + ('-O', dest)
        else:
            if check_exec('wget'):
                args = ('wget', '-c', video.url, '-O', dest)
            elif check_exec('curl'):
                args = ('curl', '-C', '-', video.url, '-o', dest)
            else:
                return 1

        self.logger.debug(' '.join(args))
        subprocess.call(args)

    def read_url(self, url):
        r = requests.get(url, stream=True, timeout=30)
        return r.iter_lines()

    def complete_download(self, text, line, *ignored):
        args = line.split(' ')
        if len(args) == 2:
            return self._complete_object()
        elif len(args) >= 3:
            return self.path_completer(args[2])

    def do_download(self, line):
        """
        download ID [FILENAME]

        Download a video

        Braces-enclosed tags are replaced with data fields. Use the 'info'
        command to see what fields are available on a given video.

        Example: download KdRRge4XYIo@youtube '{title}.{ext}'
        """
        _id, dest = self.parse_command_args(line, 2, 1)
        video = self.get_object(_id, 'get_video', ['url'])
        if not video:
            print('Video not found: %s' % _id, file=self.stderr)
            return 3

        return self.download(video, dest)

    def complete_play(self, text, line, *ignored):
        args = line.split(' ')
        if len(args) >= 2:
            return self._complete_object()

    def do_play(self, line):
        """
        play ID

        Play a video with a found player.
        """
        if not line:
            print('This command takes an argument: %s' % self.get_command_help('play', short=True), file=self.stderr)
            return 2

        ret = 0
        for _id in line.split(' '):
            video = self.get_object(_id, 'get_video', ['url'])
            error = self.play(video, _id)
            if error is not None:
                ret = error

        return ret

    def play(self, video, _id):
        if not video:
            print('Video not found: %s' % _id, file=self.stderr)
            return 3
        if not video.url:
            print('Error: the direct URL is not available.', file=self.stderr)
            return 4
        try:
            player_name = self.config.get('media_player')
            media_player_args = self.config.get('media_player_args')
            if not player_name:
                self.logger.info('You can set the media_player key to the player you prefer in the video '
                                 'configuration file.')
            self.player.play(video, player_name=player_name, player_args=media_player_args)
        except (InvalidMediaPlayer, MediaPlayerNotFound) as e:
            print('%s\nVideo URL: %s' % (e, video.url))

    def complete_info(self, text, line, *ignored):
        args = line.split(' ')
        if len(args) >= 2:
            return self._complete_object()

    def do_info(self, line):
        """
        info ID [ID2 [...]]

        Get information about a video.
        """
        if not line:
            print('This command takes an argument: %s' % self.get_command_help('info', short=True), file=self.stderr)
            return 2

        self.start_format()
        for _id in line.split(' '):
            video = self.get_object(_id, 'get_video')
            if not video:
                print('Video not found: %s' % _id, file=self.stderr)
                return 3

            self.format(video)

    def complete_playlist(self, text, line, *ignored):
        args = line.split(' ')
        if len(args) == 2:
            return ['play', 'add', 'remove', 'export', 'display', 'download']
        if len(args) >= 3:
            if args[1] in ('export', 'download'):
                return self.path_completer(args[2])
            if args[1] in ('add', 'remove'):
                return self._complete_object()

    def do_playlist(self, line):
        """
        playlist cmd [args]

        playlist add ID [ID2 ID3 ...]
        playlist remove ID [ID2 ID3 ...]
        playlist export [FILENAME]
        playlist display
        playlist download [PATH]
        playlist play
        """

        if not self.interactive:
            print('This command can be used only in interactive mode.', file=self.stderr)
            return 1

        if not line:
            print('This command takes an argument: %s' % self.get_command_help('playlist'), file=self.stderr)
            return 2

        cmd, args = self.parse_command_args(line, 2, req_n=1)
        if cmd == "add":
            _ids = args.strip().split(' ')
            for _id in _ids:
                video = self.get_object(_id, 'get_video')

                if not video:
                    print('Video not found: %s' % _id, file=self.stderr)
                    return 3

                if not video.url:
                    print('Error: the direct URL is not available.', file=self.stderr)
                    return 4

                self.PLAYLIST.append(video)
        elif cmd == "remove":
            _ids = args.strip().split(' ')
            for _id in _ids:
                video_to_remove = self.get_object(_id, 'get_video')

                if not video_to_remove:
                    print('Video not found: %s' % _id, file=self.stderr)
                    return 3

                if not video_to_remove.url:
                    print('Error: the direct URL is not available.', file=self.stderr)
                    return 4

                for video in self.PLAYLIST:
                    if video.id == video_to_remove.id:
                        self.PLAYLIST.remove(video)
                        break
        elif cmd == "export":
            filename = "playlist.m3u"
            if args:
                filename = args

            file = open(filename, 'w')
            for video in self.PLAYLIST:
                file.write('%s\r\n' % video.url)
            file.close()
        elif cmd == "display":
            for video in self.PLAYLIST:
                self.cached_format(video)
        elif cmd == "download":
            for i, video in enumerate(self.PLAYLIST):
                self.download(video, args, '%02d-{id}-{title}.{ext}' % (i+1))
        elif cmd == "play":
            for video in self.PLAYLIST:
                self.play(video, video.id)
        else:
            print('Playlist command only support "add", "remove", "display", "download" and "export" arguments.', file=self.stderr)
            return 2

    def complete_nsfw(self, text, line, begidx, endidx):
        return ['on', 'off']

    def do_nsfw(self, line):
        """
        nsfw [on | off]

        If argument is given, enable or disable the non-suitable for work behavior.

        If no argument is given, print the current behavior.
        """
        line = line.strip()
        if line:
            if line == 'on':
                self.nsfw = True
            elif line == 'off':
                self.nsfw = False
            else:
                print('Invalid argument "%s".' % line)
                return 2
        else:
            print("on" if self.nsfw else "off")

    @defaultcount()
    def do_search(self, pattern):
        """
        search PATTERN

        Search for videos matching a PATTERN.
        """
        if not pattern:
            print('This command takes an argument: %s' % self.get_command_help('search', short=True), file=self.stderr)
            return 2

        self.change_path(['search'])
        self.start_format(pattern=pattern)
        for video in self.do('search_videos', pattern=pattern, nsfw=self.nsfw):
            self.cached_format(video)



# Copyright(C) 2022 woob project
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

from typing import Dict, Callable, Tuple, Any, ClassVar
from datetime import timedelta
from dateutil import parser, tz

from woob.exceptions import (
    NeedInteractiveFor2FA, BrowserInteraction,
)
from woob.tools.date import now_as_utc
from woob.tools.value import Value

from .browsers import LoginBrowser, StatesMixin


__all__ = ["TwoFactorBrowser"]


class TwoFactorBrowser(LoginBrowser, StatesMixin):
    """
    Browser which inherits from :class:`LoginBrowser` to implement 2FA authentication.

    :param config: configuration of the backend
    :type config: :class:`BackendConfig`
    """


    TWOFA_DURATION: ClassVar[int | float | None] = None
    """
    Period to keep the same state

    It is different from STATE_DURATION which updates the expire date at each
    dump.
    """

    INTERACTIVE_NAME: ClassVar[str] = 'request_information'
    """
    Config's key which is set to a non-empty value when we are in interactive mode.
    """


    AUTHENTICATION_METHODS: ClassVar[Dict[str, Callable]] = {}
    """
    Dict of config keys and methods used for double authentication.

    Must be set up in the init to handle function pointers.
    """

    COOKIES_TO_CLEAR: ClassVar[Tuple[str, ...]] = ()
    """List of cookie keys to clear before dumping state"""

    HAS_CREDENTIALS_ONLY: ClassVar[bool] = False
    """Login can also be done with credentials without 2FA"""

    SKIP_LOCATE_BROWSER_ON_CONFIG_VALUES: ClassVar[Tuple[str, ...]] = ()
    """
    Skip locate_browser if one of the config values is defined (for example
    its useful to prevent calling twice the url that sends an OTP)
    """

    def __init__(self, config, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.config = config
        self.is_interactive = config.get(self.INTERACTIVE_NAME, Value()).get() is not None
        self.twofa_logged_date = None

    def get_expire(self) -> str | None:
        if self.STATE_DURATION is None:
            return None

        expires_dates = [now_as_utc() + timedelta(minutes=self.STATE_DURATION)]

        if self.twofa_logged_date and self.TWOFA_DURATION is not None:
            expires_dates.append(self.twofa_logged_date + timedelta(minutes=self.TWOFA_DURATION))

        return str(max(expires_dates).replace(microsecond=0))

    def dump_state(self) -> Dict[str, Any]:
        self.clear_not_2fa_cookies()
        state = super().dump_state()
        if self.twofa_logged_date:
            state['twofa_logged_date'] = str(self.twofa_logged_date)

        return state

    def should_skip_locate_browser(self) -> bool:
        for key in self.SKIP_LOCATE_BROWSER_ON_CONFIG_VALUES:
            value = self.config.get(key)
            if value is None:
                continue

            if value.get() != value.default:
                return True

        return False

    def locate_browser(self, state: Dict[str, Any]):
        if self.should_skip_locate_browser():
            return

        super().locate_browser(state)

    def load_state(self, state: Dict[str, Any]):
        super().load_state(state)
        self.twofa_logged_date = None
        if state.get('twofa_logged_date') not in (None, '', 'None'):
            twofa_logged_date = parser.parse(state['twofa_logged_date'])
            if not twofa_logged_date.tzinfo:
                twofa_logged_date = twofa_logged_date.replace(tzinfo=tz.tzlocal())
            self.twofa_logged_date = twofa_logged_date

    def init_login(self):
        """
        Abstract method to implement initiation of login on website.

        This method should raise an exception.

        SCA exceptions :
        - AppValidation for polling method
        - BrowserQuestion for SMS method, token method etc.

        Any other exceptions, default to BrowserIncorrectPassword.
        """
        raise NotImplementedError()

    def clear_init_cookies(self):
        # clear cookies to avoid some errors
        self.session.cookies.clear()

    def clear_not_2fa_cookies(self):
        # clear cookies that we don't need for 2FA
        for cookie_key in self.COOKIES_TO_CLEAR:
            if cookie_key in self.session.cookies:
                del self.session.cookies[cookie_key]

    def check_interactive(self):
        if not self.is_interactive:
            raise NeedInteractiveFor2FA()

    def do_double_authentication(self):
        """
        This method will check AUTHENTICATION_METHODS
        to dispatch to the right handle_* method.

        If no backend configuration could be found,
        it will then call init_login method.
        """

        def clear_sca_key(config_key):
            value = self.config.get(config_key)
            if value is not None:
                value.set(value.default)

        assert self.AUTHENTICATION_METHODS, 'There is no config for the double authentication.'

        for config_key, handle_method in self.AUTHENTICATION_METHODS.items():
            config_value = self.config.get(config_key, Value())
            if not config_value:
                continue

            setattr(self, config_key, config_value.get())
            if getattr(self, config_key):
                try:
                    handle_method()
                except BrowserInteraction:
                    # If a BrowserInteraction is raised during the handling of the sca_key,
                    # we need to clear it before restarting the process to prevent it to block
                    # other sca_keys handling.
                    clear_sca_key(config_key)
                    raise

                self.twofa_logged_date = now_as_utc()

                # cleaning authentication config keys
                for config_key in self.AUTHENTICATION_METHODS.keys():
                    clear_sca_key(config_key)

                break
        else:
            if not self.HAS_CREDENTIALS_ONLY:
                self.check_interactive()

            self.clear_init_cookies()
            self.init_login()

    do_login = do_double_authentication



# Copyright(C) 2012 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from woob.capabilities.housing import (CapHousing, Query, POSTS_TYPES,
                                         ADVERT_TYPES, HOUSE_TYPES)
from woob.capabilities.base import empty
from woob.tools.application.repl import ReplApplication, defaultcount
from woob.tools.application.formatters.iformatter import IFormatter, PrettyFormatter
from woob.tools.config.yamlconfig import YamlConfig


__all__ = ['AppHousing']


class HousingFormatter(IFormatter):
    MANDATORY_FIELDS = ('id', 'title', 'cost', 'currency', 'area', 'date', 'text')

    def format_obj(self, obj, alias):
        result = '%s%s%s\n' % (self.BOLD, obj.title, self.NC)
        result += 'ID: %s\n' % obj.fullid

        if hasattr(obj, 'url') and obj.url:
            result += 'URL: %s\n' % obj.url

        result += 'Cost: %s%s %s\n' % (obj.cost, obj.currency, obj.utilities)
        area = '%.2fm²' % (obj.area) if obj.area else '%s' % obj.area
        result += 'Area: %s\n' % area
        if hasattr(obj, 'price_per_meter') and not empty(obj.price_per_meter):
            result += 'Price per square meter: %.2f %s/m²\n' % (obj.price_per_meter, obj.currency)
        if hasattr(obj, 'rooms') and not empty(obj.rooms):
            result += 'Rooms: %d\n' % (obj.rooms)
        if hasattr(obj, 'bedrooms') and not empty(obj.bedrooms):
            result += 'Bedrooms: %d\n' % (obj.bedrooms)
        if obj.date:
            result += 'Date: %s\n' % obj.date.strftime('%Y-%m-%d')
        result += 'Phone: %s\n' % obj.phone
        if hasattr(obj, 'location') and obj.location:
            result += 'Location: %s\n' % obj.location
        if hasattr(obj, 'station') and obj.station:
            result += 'Station: %s\n' % obj.station

        if hasattr(obj, 'photos') and obj.photos:
            result += '\n%sPhotos%s\n' % (self.BOLD, self.NC)
            for photo in obj.photos:
                result += ' * %s\n' % photo.url

        result += '\n%sDescription%s\n' % (self.BOLD, self.NC)
        result += obj.text

        if hasattr(obj, 'details') and obj.details:
            result += '\n\n%sDetails%s\n' % (self.BOLD, self.NC)
            for key, value in obj.details.items():
                result += ' %s: %s\n' % (key, value)

        return result


class HousingListFormatter(PrettyFormatter):
    MANDATORY_FIELDS = ('id', 'title', 'cost', 'text')

    def get_title(self, obj):
        return '%s%s %s - %s' % (obj.cost, obj.currency, obj.utilities, obj.title)

    def get_description(self, obj):
        result = ''
        if hasattr(obj, 'date') and obj.date:
            result += '%s - ' % obj.date.strftime('%Y-%m-%d')
        result += obj.text
        return result


class AppHousing(ReplApplication):
    APPNAME = 'housing'
    VERSION = '3.7'
    COPYRIGHT = 'Copyright(C) 2012-YEAR Romain Bignon'
    DESCRIPTION = "Console application to search for housing."
    SHORT_DESCRIPTION = "search for housing"
    CAPS = CapHousing
    CONFIG = {'queries': {}}
    EXTRA_FORMATTERS = {'housing_list': HousingListFormatter,
                        'housing':      HousingFormatter,
                       }
    COMMANDS_FORMATTERS = {'search': 'housing_list',
                           'info': 'housing',
                           'load': 'housing_list'
                          }

    def main(self, argv):
        self.load_config(klass=YamlConfig)
        return super().main(argv)

    @defaultcount(10)
    def do_search(self, line):
        """
        search

        Search for housing. Parameters are interactively asked.
        """
        pattern = 'notempty'
        query = Query()
        query.cities = []
        while pattern:
            if len(query.cities) > 0:
                print('\n%sSelected cities:%s %s' % (self.BOLD, self.NC, ', '.join([c.name for c in query.cities])))
            pattern = self.ask('Enter a city pattern (or empty to stop)', default='')
            if not pattern:
                break

            cities = []
            for city in self.woob.do('search_city', pattern):
                cities.append(city)

            if len(cities) == 0:
                print('  Not found!')
                continue
            if len(cities) == 1:
                if city in query.cities:
                    query.cities.remove(city)
                else:
                    query.cities.append(city)
                continue

            r = 'notempty'
            while r != '':
                for i, city in enumerate(cities):
                    print('  %s%2d)%s [%s] %s (%s)' % (self.BOLD, i+1, self.NC, 'x' if city in query.cities else ' ', city.name, city.backend))
                r = self.ask('Select cities (or empty to stop)', regexp=r'(\d+|)', default='')
                if not r.isdigit():
                    continue
                r = int(r)
                if r <= 0 or r > len(cities):
                    continue
                city = cities[r-1]
                if city in query.cities:
                    query.cities.remove(city)
                else:
                    query.cities.append(city)

        r = 'notempty'
        while r != '':
            for i, good in enumerate(HOUSE_TYPES, 1):
                print('  %s%2d)%s [%s] %s' % (self.BOLD,
                                              i,
                                              self.NC,
                                              'x' if good in query.house_types else ' ', good))
            r = self.ask('Select type of house (or empty to stop)', regexp=r'(\d+|)', default='')
            if not r.isdigit():
                continue
            r = int(r)
            if r <= 0 or r > len(HOUSE_TYPES):
                continue
            value = list(HOUSE_TYPES)[r - 1]
            if value in query.house_types:
                query.house_types.remove(value)
            else:
                query.house_types.append(value)

        _type = None
        posts_types = sorted(POSTS_TYPES)
        while _type not in range(len(posts_types)):
            for i, t in enumerate(posts_types):
                print('  %s%2d)%s %s' % (self.BOLD,
                                         i,
                                         self.NC,
                                         t))
            _type = self.ask_int('Type of query')

        query.type = posts_types[_type]

        r = 'notempty'
        while r != '':
            for i, good in enumerate(ADVERT_TYPES, 1):
                print('  %s%2d)%s [%s] %s' % (self.BOLD,
                                              i,
                                              self.NC,
                                              'x' if good in query.advert_types else ' ', good))
            r = self.ask('Select type of posts (or empty to stop)', regexp=r'(\d+|)', default='')
            if not r.isdigit():
                continue
            r = int(r)
            if r <= 0 or r > len(ADVERT_TYPES):
                continue
            value = list(ADVERT_TYPES)[r - 1]
            if value in query.advert_types:
                query.advert_types.remove(value)
            else:
                query.advert_types.append(value)

        query.area_min = self.ask_int('Enter min area')
        query.area_max = self.ask_int('Enter max area')
        query.cost_min = self.ask_int('Enter min cost')
        query.cost_max = self.ask_int('Enter max cost')
        query.nb_rooms = self.ask_int('Enter number of rooms')
        save_query = self.ask('Save query (y/n)', default='n')
        if save_query.upper() == 'Y':
            name = ''
            while not name:
                name = self.ask('Query name')

            self.config.set('queries', name, query)
            self.config.save()
        self.complete_search(query)

    def complete_search(self, query):
        self.change_path(['housings'])
        self.start_format()
        for housing in self.do('search_housings', query):
            self.cached_format(housing)

    def ask_int(self, txt):
        r = self.ask(txt, default='', regexp=r'(\d+|)')
        if r:
            return int(r)
        return None

    @defaultcount(10)
    def do_load(self, query_name):
        """
        load [query name]
        without query name : list loadable queries
        with query name laod query
        """
        queries = self.config.get('queries')
        if not queries:
            print('There is no saved queries', file=self.stderr)
            return 2

        if not query_name:
            for name in queries.keys():
                print('  %s* %s %s' % (self.BOLD,
                                       self.NC,
                                       name))
            query_name = self.ask('Which one')

        if query_name in queries:
            self.complete_search(queries.get(query_name))
        else:
            print('Unknown query', file=self.stderr)
            return 2

    def complete_info(self, text, line, *ignored):
        args = line.split(' ')
        if len(args) == 2:
            return self._complete_object()

    def do_info(self, _id):
        """
        info ID

        Get information about a housing.
        """
        if not _id:
            print('This command takes an argument: %s' % self.get_command_help('info', short=True), file=self.stderr)
            return 2

        housing = self.get_object(_id, 'get_housing')
        if not housing:
            print('Housing not found: %s' % _id, file=self.stderr)
            return 3

        self.start_format()
        self.format(housing)



# Copyright(C) 2010-2011 Romain Bignon, Christophe Benz, Noé Rubinstein
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.


from woob.capabilities.image import BaseImage as CIBaseImage, Thumbnail
from .base import Capability, BaseObject, NotLoaded, Field, StringField, \
                  IntField, FloatField, Enum
from .date import DateField


__all__ = ['BaseGallery', 'BaseImage', 'CapGallery']


class BaseGallery(BaseObject):
    """
    Represents a gallery.

    This object has to be inherited to specify how to calculate the URL of the gallery from its ID.
    """
    title =         StringField('Title of gallery')
    description =   StringField('Description of gallery')
    cardinality =   IntField('Cardinality of gallery')
    date =          DateField('Date of gallery')
    rating =        FloatField('Rating of this gallery')
    rating_max =    FloatField('Max rating available')
    thumbnail =     Field('Thumbnail', Thumbnail)

    def __init__(self, _id, title=NotLoaded, url=NotLoaded, cardinality=NotLoaded, date=NotLoaded,
                 rating=NotLoaded, rating_max=NotLoaded, thumbnail=NotLoaded, thumbnail_url=None, nsfw=False):
        super(BaseGallery, self).__init__(str(_id), url)

        self.title = title
        self.date = date
        self.rating = rating
        self.rating_max = rating_max
        self.thumbnail = thumbnail

    @classmethod
    def id2url(cls, _id):
        """Overloaded in child classes provided by backends."""
        raise NotImplementedError()

    @property
    def page_url(self):
        """
        Get URL to page of this gallery.
        """
        return self.id2url(self.id)

    def iter_image(self):
        """
        Iter images.
        """
        raise NotImplementedError()


class BaseImage(CIBaseImage):
    """
    Base class for images.
    """
    index =     IntField('Usually page number')
    gallery =   Field('Reference to the Gallery object', BaseGallery)

    def __init__(self, _id='', index=None, thumbnail=NotLoaded, url=NotLoaded,
            ext=NotLoaded, gallery=None):

        super(BaseImage, self).__init__(str(_id), url)

        self.index = index
        self.thumbnail = thumbnail
        self.ext = ext
        self.gallery = gallery

    def __str__(self):
        return self.url

    def __repr__(self):
        return '<Image url="%s">' % self.url

    def __iscomplete__(self):
        return self.data is not NotLoaded


class SearchSort(Enum):
    RELEVANCE = 0
    RATING = 1
    VIEWS = 2
    DATE = 3


class CapGallery(Capability):
    """
    This capability represents the ability for a website backend to provide videos.
    """
    SEARCH_RELEVANCE = SearchSort.RELEVANCE
    SEARCH_RATING = SearchSort.RATING
    SEARCH_VIEWS = SearchSort.VIEWS
    SEARCH_DATE = SearchSort.DATE

    def search_galleries(self, pattern, sortby=SEARCH_RELEVANCE):
        """
        Iter results of a search on a pattern.

        :param pattern: pattern to search on
        :type pattern: str
        :param sortby: sort by...
        :type sortby: SEARCH_*
        :rtype: :class:`BaseGallery`
        """
        raise NotImplementedError()

    def get_gallery(self, _id):
        """
        Get gallery from an ID.

        :param _id: the gallery id. It can be a numeric ID, or a page url, or so.
        :type _id: str
        :rtype: :class:`Gallery`
        """
        raise NotImplementedError()

    def iter_gallery_images(self, gallery):
        """
        Iterate images from a Gallery.

        :type gallery: BaseGallery
        :rtype: iter(BaseImage)
        """
        raise NotImplementedError()



# Copyright(C) 2010-2011 Christophe Benz, Romain Bignon, John Obbele
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.


from contextlib import closing
from subprocess import PIPE, Popen
import subprocess
import shlex
from shutil import which

import requests

from woob.tools.log import getLogger

__all__ = ['InvalidMediaPlayer', 'MediaPlayer', 'MediaPlayerNotFound']


PLAYERS = (
    ('mpv',      '-'),
    ('mplayer2', '-'),
    ('mplayer',  '-'),
    ('vlc',      '-'),
    ('parole',   'fd://0'),
    ('totem',    'fd://0'),
    ('xine',     'stdin:/'),
)


class MediaPlayerNotFound(Exception):
    def __init__(self):
        super(MediaPlayerNotFound, self).__init__('No media player found on this system. Please install one of them: %s.' %
                           ', '.join(player[0] for player in PLAYERS))


class InvalidMediaPlayer(Exception):
    def __init__(self, player_name):
        super(InvalidMediaPlayer, self).__init__('Invalid media player: %s. Valid media players: %s.' % (
            player_name, ', '.join(player[0] for player in PLAYERS)))


class MediaPlayer:
    """
    Black magic invoking a media player to this world.

    Presently, due to strong disturbances in the holidays of the ether
    world, the media player used is chosen from a static list of
    programs. See PLAYERS for more information.
    """

    def __init__(self, logger=None):
        if logger:
            self.logger = getLogger("mediaplayer", logger)
        else:
            self.logger = getLogger(__name__)

    def guess_player_name(self):
        for player_name in [player[0] for player in PLAYERS]:
            if which(player_name) is not None:
                return player_name
        return None

    def play(self, media, player_name=None, player_args=None):
        """
        Play a media object, using programs from the PLAYERS list.

        This function dispatch calls to either _play_default or
        _play_rtmp for special rtmp streams using SWF verification.
        """
        player_names = [player[0] for player in PLAYERS]
        if not player_name:
            self.logger.debug('No media player given. Using the first available from: %s.' %
                              ', '.join(player_names))
            player_name = self.guess_player_name()
            if player_name is None:
                raise MediaPlayerNotFound()
        if media.url.startswith('rtmp'):
            self._play_rtmp(media, player_name, args=player_args)
        else:
            self._play_default(media, player_name, args=player_args)

    def _play_default(self, media, player_name, args=None):
        """
        Play media.url with the media player.
        """
        # if flag play_proxy...
        if hasattr(media, '_play_proxy') and media._play_proxy is True:
            # use requests to handle redirect and cookies
            self._play_proxy(media, player_name, args)
            return None

        args = shlex.split(player_name)
        args.append(media.url)

        print('Invoking "%s".' % (' '.join(args)))
        subprocess.call(args)

    def _play_proxy(self, media, player_name, args):
        """
        Load data with python requests and pipe data to a media player.

        We need this function for url that use redirection and cookies.
        This function is used if the non-standard,
        non-API compliant '_play_proxy' attribute of the 'media' object is defined and is True.
        """
        if args is None:
            for (binary, stdin_args) in PLAYERS:
                if binary == player_name:
                    args = stdin_args

        assert args is not None

        print(':: Play_proxy streaming from %s' % media.url)
        print(':: to %s %s' % (player_name, args))
        print(player_name + ' ' + args)
        cmd = shlex.split(player_name) + shlex.split(args)
        proc = Popen(cmd, stdin=PIPE)

        # Handle cookies (and redirection 302...)
        session = requests.sessions.Session()

        with closing(proc.stdin):
            with closing(session.get(media.url, stream=True)) as response:
                for buffer in response.iter_content(8192):
                    try:
                        proc.stdin.write(buffer)
                    except:
                        print("play_proxy broken pipe. Can't write anymore.")
                        break

    def _play_rtmp(self, media, player_name, args):
        """
        Download data with rtmpdump and pipe them to a media player.

        You need a working version of rtmpdump installed and the SWF
        object url in order to comply with SWF verification requests
        from the server. The last one is retrieved from the non-standard
        non-API compliant 'swf_player' attribute of the 'media' object.
        """
        if which('rtmpdump') is None:
            self.logger.warning('"rtmpdump" binary not found')
            return self._play_default(media, player_name)
        media_url = media.url
        try:
            player_url = media.swf_player
            if media.swf_player:
                rtmp = 'rtmpdump -r %s --swfVfy %s' % (media_url, player_url)
            else:
                rtmp = 'rtmpdump -r %s' % media_url
        except AttributeError:
            self.logger.warning('Your media object does not have a "swf_player" attribute. SWF verification will be '
                                'disabled and may prevent correct media playback.')
            return self._play_default(media, player_name)

        rtmp += ' --quiet'

        if args is None:
            for (binary, stdin_args) in PLAYERS:
                if binary == player_name:
                    args = stdin_args

        assert args is not None

        player_name = shlex.split(player_name)
        args = shlex.split(args)

        print(':: Streaming from %s' % media_url)
        print(':: to %s %s' % (player_name, args))
        print(':: %s' % rtmp)
        p1 = Popen(shlex.split(rtmp), stdout=PIPE)
        Popen(player_name + args, stdin=p1.stdout, stderr=PIPE)



# Copyright(C) 2010-2011 Romain Bignon, Christophe Benz, Noé Rubinstein
# Copyright(C) 2013 Pierre Mazière
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from collections import OrderedDict

from .base import empty, Field, BoolField, BytesField
from .file import CapFile, BaseFile

__all__ = ['BaseImage', 'Thumbnail', 'CapImage']


class _BaseImage(BaseFile):
    """
    Fake class to allow the inclusion of a BaseImage property within
    the real BaseImage class
    """
    pass


class Thumbnail(_BaseImage):
    """
    Thumbnail of an image.
    """

    data = BytesField('Data')

    def __init__(self, url=""):
        super(Thumbnail, self).__init__(url)
        self.url = url.replace(' ', '%20')

    def __str__(self):
        return self.url

    def __repr__(self):
        return '<Thumbnail url=%r>' % self.url

    def __iscomplete__(self):
        return not empty(self.data)


class BaseImage(_BaseImage):
    """
    Represents an image file.
    """
    nsfw =      BoolField('Is this Not Safe For Work', default=False)
    thumbnail = Field('Thumbnail of the image', Thumbnail)
    data =      BytesField('Data of image')

    def __iscomplete__(self):
        return not empty(self.data)

    def to_dict(self):
        def iter_decorate(d):
            for key, value in d:
                if key == 'data':
                    continue

                if key == 'id' and self.backend is not None:
                    value = self.fullid

                yield key, value

        fields_iterator = self.iter_fields()
        return OrderedDict(iter_decorate(fields_iterator))


class CapImage(CapFile):
    """
    Image file provider
    """

    def search_image(self, pattern, sortby=CapFile.SEARCH_RELEVANCE, nsfw=False):
        """
        search for an image file

        :param pattern: pattern to search on
        :type pattern: str
        :param sortby: sort by ...(use SEARCH_* constants)
        :param nsfw: include non-suitable for work images if True
        :type nsfw: bool
        :rtype: iter[:class:`BaseImage`]
        """
        return self.search_file(pattern, sortby)

    def get_image(self, _id):
        """
        Get an image file from an ID.

        :param id: image file ID
        :type id: str
        :rtype: :class:`BaseImage`]
        """
        return self.get_file(_id)



# Copyright(C) 2017-2021 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.


import multiprocessing
import os

from .util import time_buffer


__all__ = ['AutoCleanConfig', 'ForkingConfig', 'TimeBufferConfig']


"""
These classes add functionality to existing IConfig classes.
Example:

    class MyYamlConfig(TimeBufferConfig, ForkingConfig, YamlConfig):
        saved_since_seconds = 42

The recommended order is TimeBufferConfig, AutoCleanConfig, ForkingConfig, and then the
actual storage class.
"""


class AutoCleanConfig:
    """
    Removes config file if it has no values.
    """
    def save(self):
        if self.values:
            super(AutoCleanConfig, self).save()
        else:
            try:
                os.remove(self.path)
            except OSError:
                pass


class ForkingConfig:
    """
    Runs the actual save in a forked processes, making save non-blocking.
    It prevents two save() from being called at once by blocking on the previous one
    if it is not finished.
    It is also possible to call join() to wait for the save to complete.
    """
    process = None

    def __init__(self, *args, **kwargs):
        self.lock = multiprocessing.RLock()
        super(ForkingConfig, self).__init__(*args, **kwargs)

    def join(self):
        with self.lock:
            if self.process:
                self.process.join()
            self.process = None

    def save(self):
        # if a save is already in progress, wait for it to finish
        self.join()

        parent_save = super(ForkingConfig, self).save
        with self.lock:
            self.process = multiprocessing.Process(target=parent_save, name='save %s' % self.path)
            self.process.start()

    def __exit__(self, t, v, tb):
        self.join()
        super(ForkingConfig, self).__exit__(t, v, tb)

    def __getstate__(self):
        d = self.__dict__.copy()
        d.pop('lock', None)
        return d

    def __setstate__(self, d):
        self.__init__(path=d['path'])
        for k, v in d.items():
            setattr(self, k, v)


class TimeBufferConfig:
    """
    Really saves only every saved_since_seconds seconds.
    It is possible to force save (e.g. at exit) with force_save().
    """
    saved_since_seconds = None

    def __init__(self, path, saved_since_seconds=None, last_run=True, logger=None, *args, **kwargs):
        super(TimeBufferConfig, self).__init__(path, *args, **kwargs)
        if saved_since_seconds:
            self.saved_since_seconds = saved_since_seconds
        if self.saved_since_seconds:
            self.save = time_buffer(since_seconds=self.saved_since_seconds, last_run=last_run, logger=logger)(self.save)

    def save(self, *args, **kwargs):
        kwargs.pop('since_seconds', None)
        super(TimeBufferConfig, self).save(*args, **kwargs)

    def force_save(self):
        self.save(since_seconds=False)

    def __exit__(self, t, v, tb):
        self.force_save()
        super(TimeBufferConfig, self).__exit__(t, v, tb)

    def __getstate__(self):
        try:
            d = super(TimeBufferConfig, self).__getstate__()
        except AttributeError:
            d = self.__dict__.copy()
        # When decorated, it is not serializable.
        # The decorator will be added again by __setstate__.
        d.pop('save', None)
        return d

    def __setstate__(self, d):
        # Add the decorator if needed
        self.__init__(path=d['path'],
                      saved_since_seconds=d.get('saved_since_seconds'))
        for k, v in d.items():
            setattr(self, k, v)



# Copyright(C) 2010-2012  Romain Bignon, Florent Fourcot
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.


from .base import Capability, BaseObject, StringField, DecimalField, Field, UserError, empty
from .date import DateField
from .address import compat_field, GeoCoordinates, PostalAddress

__all__ = ['Gauge', 'GaugeSensor', 'GaugeMeasure', 'CapGauge', 'SensorNotFound']


class SensorNotFound(UserError):
    """
    Not found a sensor
    """


class Gauge(BaseObject):
    """
    Gauge class.
    """
    name =       StringField('Name of gauge')
    city =       StringField('City of the gauge')
    object =     StringField('What is evaluate')  # For example, name of a river
    sensors =    Field('List of sensors on the gauge', list)


class GaugeMeasure(BaseObject):
    """
    Measure of a gauge sensor.
    """
    level =     DecimalField('Level of measure')
    date =      DateField('Date of measure')
    alarm =     StringField('Alarm level')

    def __repr__(self):
        if empty(self.level):
            return "<GaugeMeasure is %s>" % self.level
        else:
            return "<GaugeMeasure level=%f alarm=%s date=%s>" % (self.level, self.alarm, self.date)


class GaugeSensor(BaseObject):
    """
    GaugeSensor class.
    """
    name =      StringField('Name of the sensor')
    unit =      StringField('Unit of values')
    forecast =  StringField('Forecast')
    location = Field('Address of the sensor', PostalAddress)
    geo = Field('Geo address of the sensor', GeoCoordinates)

    address = compat_field('location', 'full_address')
    latitude = compat_field('geo', 'latitude')
    longitude = compat_field('geo', 'longitude')

    lastvalue = Field('Last value', GaugeMeasure)
    history =   Field('Value history', list)  # lastvalue not included
    gaugeid =   StringField('Id of the gauge')

    def __repr__(self):
        return "<GaugeSensor id=%s name=%s>" % (self.id, self.name)


class CapGauge(Capability):
    def iter_gauges(self, pattern=None):
        """
        Iter gauges.

        :param pattern: if specified, used to search gauges.
        :type pattern: str
        :rtype: iter[:class:`Gauge`]
        """
        raise NotImplementedError()

    def iter_sensors(self, id, pattern=None):
        """
        Iter instrument of a gauge.

        :param: ID of the gauge
        :param pattern: if specified, used to search sensors.
        :type pattern: str
        :rtype: iter[:class:`GaugeSensor`]
        """
        raise NotImplementedError()

    def iter_gauge_history(self, id):
        """
        Get history of a gauge sensor.

        :param id: ID of the gauge sensor
        :type id: str
        :rtype: iter[:class:`GaugeMeasure`]
        """
        raise NotImplementedError()

    def get_last_measure(self, id):
        """
        Get last measures of a sensor.

        :param id: ID of the sensor.
        :type id: str
        :rtype: :class:`GaugeMeasure`
        """
        raise NotImplementedError()



# Copyright(C) 2019-2021 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.


import os
import sqlite3
import tempfile
from collections.abc import Mapping, MutableMapping

import yaml

from .iconfig import ConfigError, IConfig
from .util import replace, time_buffer
from .yamlconfig import WoobDumper

try:
    from yaml import CSafeLoader as SafeLoader
except ImportError:
    from yaml import SafeLoader


__all__ = ['SQLiteConfig']


def quote(s, errors="strict"):
    encodable = s.encode("utf-8", errors).decode("utf-8")
    nul_index = encodable.find("\x00")
    if nul_index >= 0:
        raise ValueError("invalid identifier")

    return "\"" + encodable.replace("\"", "\"\"") + "\""


class VirtualRootDict(Mapping):
    def __init__(self, config):
        self.config = config

    def __getitem__(self, base):
        if base in self.config._tables:
            return VirtualDict(self.config, base)
        raise KeyError('%s table not found' % base)

    def __iter__(self):
        for base in self.config._tables:
            yield base

    def __len__(self):
        return len(self.config._tables)


class VirtualDict(MutableMapping):
    def __init__(self, config, base):
        self.config = config
        self.base = base

    def __getitem__(self, key):
        try:
            return self.config.get(self.base, key)
        except ConfigError:
            raise KeyError('%s key in %s table not found' % (key, self.base))

    def __contains__(self, key):
        return self.config.has(self.base, key)

    def __iter__(self):
        for key in self.config.keys(self.base):
            yield key

    def items(self):
        return self.config.items(self.base)

    def __len__(self):
        return self.config.count(self.base)

    def __delitem__(self, key):
        try:
            self.config.delete(self, self.base, key)
        except ConfigError:
            raise KeyError('%s key in %s table not found' % (key, self.base))

    def __setitem__(self, key, value):
        self.config.set(self.base, key, value)


class SQLiteConfig(IConfig):
    commit_since_seconds = 3600
    dump_since_seconds = 600

    def __init__(self, path, commit_since_seconds=None, dump_since_seconds=None, last_run=True, logger=None):
        self.path = path
        if commit_since_seconds:
            self.commit_since_seconds = commit_since_seconds
        if dump_since_seconds:
            self.dump_since_seconds = dump_since_seconds
        if self.commit_since_seconds:
            self.commit = time_buffer(since_seconds=self.commit_since_seconds, last_run=last_run, logger=logger)(self.commit)
        if self.dump_since_seconds:
            self.dump = time_buffer(since_seconds=self.dump_since_seconds, last_run=last_run, logger=logger)(self.dump)

    def load(self, default={}, optimize=True):
        self.storage = sqlite3.connect(self.path)
        self.storage.execute('PRAGMA page_size = 4096')
        if optimize:
            self.storage.execute('VACUUM')
            self.storage.execute('REINDEX')
        self._tables = set(self.tables())
        self.values = VirtualRootDict(self)

    def save(self, commit_since_seconds=None, dump_since_seconds=None):
        self.commit(since_seconds=commit_since_seconds)
        # No one would want immediate dumps, assume it means no dumps
        if self.dump_since_seconds:
            self.dump(since_seconds=dump_since_seconds)

    def force_save(self):
        self.save(commit_since_seconds=False, dump_since_seconds=False)

    def __exit__(self, t, v, tb):
        self.force_save()
        super(SQLiteConfig, self).__exit__(t, v, tb)

    def commit(self, **kwargs):
        kwargs.pop('since_seconds', None)
        self.storage.commit()

    def dump(self, **kwargs):
        kwargs.pop('since_seconds', None)
        target = os.path.splitext(self.path)[0] + '.sql'
        with tempfile.NamedTemporaryFile(dir=os.path.dirname(self.path), delete=False) as f:
            for line in self.storage.iterdump():
                f.write(str(line).encode('utf-8'))
                f.write(b'\n')
        replace(f.name, target)

    def ensure_table(self, name):
        if name not in self._tables:
            self.storage.execute('''CREATE TABLE IF NOT EXISTS %s (
                key text PRIMARY KEY,
                value text
            );''' % quote(name))  # nosec
            self._tables.add(name)

    def tables(self):
        cur = self.storage.cursor()
        cur.execute('''SELECT name FROM sqlite_master
            WHERE type="table" AND name NOT LIKE "sqlite_%";''')
        return [k[0] for k in cur.fetchall()]

    def items(self, table, size=100):
        """
        Low memory way of listing all items.
        The size parameters alters how many items are fetched at a time.
        """
        cur = self.storage.cursor()
        cur.execute('SELECT key, value FROM %s;' % quote(table))  # nosec
        items = cur.fetchmany(size)
        while items:
            for key, strvalue in items:
                yield key, yaml.load(strvalue, Loader=SafeLoader)
            items = cur.fetchmany(size)

    def keys(self, table, size=200):
        """
        Low memory way of listing all keys.
        The size parameters alters how many items are fetched at a time.
        """
        cur = self.storage.cursor()
        cur.execute('SELECT key FROM %s;' % quote(table))  # nosec
        items = cur.fetchmany(size)
        while items:
            for item in items:
                yield item[0]
            items = cur.fetchmany(size)

    def count(self, table):
        cur = self.storage.cursor()
        cur.execute('SELECT count(*) FROM %s;' % quote(table))  # nosec
        return cur.fetchone()[0]

    def get(self, *args, **kwargs):
        table = args[0]
        key = '.'.join(args[1:])
        self.ensure_table(table)
        if not key:
            return self.values[table]
        try:
            cur = self.storage.cursor()
            cur.execute('SELECT value FROM %s WHERE key=?;' % quote(table), (key, ))  # nosec
            row = cur.fetchone()
            if row is None:
                if 'default' in kwargs:
                    value = kwargs.get('default')
                else:
                    raise ConfigError()
            else:
                strvalue = row[0]
                value = yaml.load(strvalue, Loader=SafeLoader)
        except TypeError:
            raise ConfigError()
        return value

    def set(self, *args):
        table = args[0]
        key = '.'.join(args[1:-1])
        if not key:
            raise ConfigError('A minimum of two levels are required.')
        value = args[-1]
        self.ensure_table(table)
        try:
            strvalue = yaml.dump(value, None, Dumper=WoobDumper, default_flow_style=False)
            cur = self.storage.cursor()
            cur.execute('''INSERT OR REPLACE INTO %s VALUES (?, ?)''' % quote(table), (key, strvalue))  # nosec
        except KeyError:
            raise ConfigError()
        except TypeError:
            raise ConfigError()

    def delete(self, *args):
        table = args[0]
        key = '.'.join(args[1:])
        if not key:
            if table in self._tables:
                cur = self.storage.cursor()
                cur.execute('DROP TABLE %s;' % quote(table))  # nosec
            else:
                raise ConfigError()
        else:
            self.ensure_table(table)
            cur = self.storage.cursor()
            cur.execute('DELETE FROM %s WHERE key=?;' % quote(table), (key, ))  # nosec
            if not cur.rowcount:
                raise ConfigError()

    def has(self, *args):
        table = args[0]
        key = '.'.join(args[1:])
        if not key:
            return table in self._tables
        cur = self.storage.cursor()
        cur.execute('SELECT count(*) FROM %s WHERE key=?;' % quote(table), (key, ))  # nosec
        return cur.fetchone()[0] > 0



# Copyright(C) 2010-2021  Noé Rubinstein
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

import os
from re import search, sub

from woob.tools.application.repl import ReplApplication, defaultcount
from woob.capabilities.base import empty
from woob.capabilities.gallery import CapGallery, BaseGallery, BaseImage
from woob.tools.application.formatters.iformatter import PrettyFormatter


__all__ = ['AppGallery']


class GalleryListFormatter(PrettyFormatter):
    MANDATORY_FIELDS = ('id', 'title')

    def get_title(self, obj):
        s = obj.title
        if hasattr(obj, 'cardinality') and not empty(obj.cardinality):
            s += ' (%d pages)' % obj.cardinality
        return s

    def get_description(self, obj):
        if hasattr(obj, 'description') and obj.description:
            return obj.description


class AppGallery(ReplApplication):
    APPNAME = 'gallery'
    VERSION = '3.7'
    COPYRIGHT = 'Copyright(C) 2011-YEAR Noé Rubinstein'
    DESCRIPTION = 'gallery browses and downloads web image galleries'
    SHORT_DESCRIPTION = 'browse and download web image galleries'
    CAPS = CapGallery
    EXTRA_FORMATTERS = {'gallery_list': GalleryListFormatter}
    COMMANDS_FORMATTERS = {'search': 'gallery_list', 'ls': 'gallery_list'}
    COLLECTION_OBJECTS = (BaseGallery, BaseImage, )

    def __init__(self, *args, **kwargs):
        super(AppGallery, self).__init__(*args, **kwargs)

    @defaultcount(10)
    def do_search(self, pattern):
        """
        search PATTERN

        List galleries matching a PATTERN.
        """
        if not pattern:
            print('This command takes an argument: %s' % self.get_command_help('search', short=True), file=self.stderr)
            return 2

        self.start_format(pattern=pattern)
        for gallery in self.do('search_galleries', pattern=pattern):
            self.cached_format(gallery)

    def do_download(self, line):
        """
        download ID [FIRST [FOLDER]]

        Download a gallery.

        Begins at page FIRST (default: 0) and saves to FOLDER (default: title)
        """
        _id, first, dest = self.parse_command_args(line, 3, 1)

        if first is None:
            first = 0
        else:
            first = int(first)

        gallery = None
        _id, backend = self.parse_id(_id)
        for result in self.do('get_gallery', _id, backends=backend):
            if result:
                gallery = result

        if not gallery:
            print('Gallery not found: %s' % _id, file=self.stderr)
            return 3

        self.woob[backend].fillobj(gallery, ('title',))
        if dest is None:
            dest = sub('/', ' ', gallery.title)

        print("Downloading to %s" % dest)

        try:
            os.mkdir(dest)
        except OSError:
            pass  # ignore error on existing directory
        os.chdir(dest)  # fail here if dest couldn't be created

        i = 0
        for img in self.woob[backend].iter_gallery_images(gallery):
            i += 1
            if i < first:
                continue

            self.woob[backend].fillobj(img, ('url', 'data'))
            if img.data is None:
                self.woob[backend].fillobj(img, ('url', 'data'))
                if img.data is None:
                    print("Couldn't get page %d, exiting" % i, file=self.stderr)
                    break

            ext = search(r"\.([^\.]{1,5})$", img.url)
            if ext:
                ext = ext.group(1)
            else:
                ext = "jpg"

            name = '%03d.%s' % (i, ext)
            print('Writing file %s' % name)

            with open(name, 'wb') as f:
                f.write(img.data)

        os.chdir(os.path.pardir)

    def do_info(self, line):
        """
        info ID

        Get information about a gallery.
        """
        _id, = self.parse_command_args(line, 1, 1)

        gallery = self.get_object(_id, 'get_gallery')
        if not gallery:
            print('Gallery not found: %s' % _id, file=self.stderr)
            return 3

        self.start_format()
        self.format(gallery)



# flake8: compatible

from __future__ import annotations

import shlex

import requests

from woob.tools.json import json

__all__ = ['to_curl']


def to_curl(request: requests.PreparedRequest | dict) -> str:
    """Return a generated functional curl command based on a request.

    :param request: The prepared request, or property dictionary, to
        transform into a curl command.
    :return: The curl command, with UNIX newlines.
    """

    if isinstance(request, requests.PreparedRequest):
        method: str = request.method
        url: str = request.url
        headers: dict[str, str] = request.headers
        body: bytes | None = request.body
    else:
        method = request['method']
        url = request['url']
        headers = json.loads(str(request['headers']).replace("'", '"'))
        body = request.get('body')

    parts = [
        'curl',
        '--compressed',  # Decompress encoded data before stdout
    ]

    if method not in ('GET', 'POST'):
        parts += ('-X', method)

    for header, value in headers.items():
        parts += ['-H', f'{header}:{value}']

    if body:
        if isinstance(body, bytes):
            parts += ['-d', f'{body.decode("utf-8")}']
        else:
            parts += ['-d', f'{body}']

    parts += [f"{url}"]

    return shlex.join(parts)



# Copyright(C) 2010-2014 Romain Bignon, Christophe Benz
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.


from copy import copy
from threading import Thread, Event
import queue

from woob.capabilities.base import BaseObject
from woob.tools.misc import get_backtrace
from woob.tools.log import getLogger


__all__ = ['BackendsCall', 'CallErrors']


class CallErrors(Exception):
    def __init__(self, errors):
        msg = 'Errors during backend calls:\n' + \
                '\n'.join(['Module(%r): %r\n%r\n' % (backend, error, backtrace)
                           for backend, error, backtrace in errors])

        super(CallErrors, self).__init__(msg)
        self.errors = copy(errors)

    def __iter__(self):
        return self.errors.__iter__()


class BackendsCall:
    def __init__(self, backends, function, *args, **kwargs):
        """
        :param backends: List of backends to call
        :type backends: list[:class:`Module`]
        :param function: backends' method name, or callable object.
        :type function: :class:`str` or :class:`callable`
        """
        self.logger = getLogger(__name__)

        self.responses = queue.Queue()
        self.errors = []
        self.tasks = queue.Queue()
        self.stop_event = Event()
        self.threads = []

        for backend in backends:
            t = Thread(target=self.backend_process, args=(function, args, kwargs))
            t.start()
            self.threads.append(t)
            self.tasks.put(backend)

    def store_result(self, backend, result):
        """Store the result when a backend task finished."""
        if result is None:
            return

        if isinstance(result, BaseObject):
            result.backend = backend.name
        self.responses.put(result)

    def backend_process(self, function, args, kwargs):
        """
        Internal method to run a method of a backend.

        As this method may be blocking, it should be run on its own thread.
        """
        backend = self.tasks.get()
        with backend:
            try:
                # Call method on backend
                try:
                    self.logger.debug('%s: Calling function %s', backend, function)
                    if callable(function):
                        result = function(backend, *args, **kwargs)
                    else:
                        result = getattr(backend, function)(*args, **kwargs)
                except Exception as error:
                    self.logger.debug('%s: Called function %s raised an error: %r', backend, function, error)
                    self.errors.append((backend, error, get_backtrace(error)))
                else:
                    self.logger.debug('%s: Called function %s returned: %r', backend, function, result)

                    if hasattr(result, '__iter__') and not isinstance(result, (bytes, str)):
                        # Loop on iterator
                        try:
                            for subresult in result:
                                self.store_result(backend, subresult)
                                if self.stop_event.is_set():
                                    break
                        except Exception as error:
                            self.errors.append((backend, error, get_backtrace(error)))
                    else:
                        self.store_result(backend, result)
            finally:
                self.tasks.task_done()

    def _callback_thread_run(self, callback, errback, finishback):
        while not self.stop_event.is_set() and (self.tasks.unfinished_tasks or not self.responses.empty()):
            try:
                response = self.responses.get(timeout=0.1)
            except queue.Empty:
                continue
            else:
                if callback:
                    callback(response)

        # Raise errors
        while errback and self.errors:
            errback(*self.errors.pop(0))

        if finishback:
            finishback()

    def callback_thread(self, callback, errback=None, finishback=None):
        """
        Call this method to create a thread which will callback a
        specified function everytimes a new result comes.

        When the process is over, the function will be called with
        both arguments set to None.

        The functions prototypes:
            def callback(result)
            def errback(backend, error, backtrace)
            def finishback()

        """
        thread = Thread(target=self._callback_thread_run, args=(callback, errback, finishback))
        thread.start()
        return thread

    def wait(self):
        """Wait until all tasks are finished."""
        for thread in self.threads:
            thread.join()

        if self.errors:
            raise CallErrors(self.errors)

    def stop(self, wait=False):
        """
        Stop all tasks.

        :param wait: If True, wait until all tasks stopped.
        :type wait: bool
        """

        self.stop_event.set()

        if wait:
            self.wait()

    def __iter__(self):
        try:
            while not self.stop_event.is_set() and (self.tasks.unfinished_tasks or not self.responses.empty()):
                try:
                    yield self.responses.get(timeout=0.1)
                except queue.Empty:
                    continue
        except:
            self.stop()
            raise

        if self.errors:
            raise CallErrors(self.errors)



# Copyright(C) 2010-2011 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from woob.tools.application.repl import ReplApplication


__all__ = ['AppCli']


class AppCli(ReplApplication):
    APPNAME = 'cli'
    VERSION = '3.7'
    COPYRIGHT = 'Copyright(C) 2010-YEAR Romain Bignon'
    SYNOPSIS =  'Usage: %prog [-dqv] [-b backends] [-cnfs] capability method [arguments..]\n'
    SYNOPSIS += '       %prog [--help] [--version]'
    DESCRIPTION = "Console application to call a specific method on backends " \
                  "which implement the given capability."
    SHORT_DESCRIPTION = "call a method on backends"
    DISABLE_REPL = True

    def load_default_backends(self):
        pass

    def main(self, argv):
        if len(argv) < 3:
            print("Syntax: %s capability method [args ..]" % argv[0], file=self.stderr)
            return 2

        cap_s = argv[1]
        cmd = argv[2]
        args = argv[3:]

        self.load_backends(cap_s)

        for obj in self.do(cmd, *args):
            self.format(obj)

        self.flush()

        return 0



# Copyright(C) 2010-2014 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

from time import time, sleep
from typing import Optional, Union
import itertools
import locale
import os
import re
import sys
import traceback
import types
import unicodedata
import unidecode

__all__ = [
    'NO_DEFAULT',
    'NoDefaultType',
    'classproperty',
    'clean_text',
    'find_exe',
    'get_backtrace',
    'get_bytes_size',
    'iter_fields',
    'limit',
    'polling_loop',
    'to_unicode',
]

NEWLINES_RE = re.compile(r'\s+', flags=re.UNICODE)


class NoDefaultType:
    """Type for the NO_DEFAULT constant."""

    __slots__ = ()

    def __new__(mcls, *args, **kwargs):
        try:
            return mcls.__unique_value__
        except AttributeError:
            value = super().__new__(mcls, *args, **kwargs)
            mcls.__unique_value__ = value
            return value

    def __repr__(self):
        return 'NO_DEFAULT'


NO_DEFAULT = NoDefaultType()


def get_backtrace(empty="Empty backtrace."):
    """
    Try to get backtrace as string.
    Returns "Error while trying to get backtrace" on failure.
    """
    try:
        info = sys.exc_info()
        if info != (None, None, None):
            trace = traceback.format_exception(*info)
            return "".join(trace)
    except:
        return "Error while trying to get backtrace"
    return empty


def get_bytes_size(size, unit_name):
    r"""Converts a unit and a number into a number of bytes.

    >>> get_bytes_size(2, 'KB')
    2048.0
    """
    unit_data = {
        'bytes': 1,
        'KB': 1024,
        'KiB': 1024,
        'MB': 1024 * 1024,
        'MiB': 1024 * 1024,
        'GB': 1024 * 1024 * 1024,
        'GiB': 1024 * 1024 * 1024,
        'TB': 1024 * 1024 * 1024 * 1024,
        'TiB': 1024 * 1024 * 1024 * 1024,
        }
    return float(size * unit_data.get(unit_name, 1))


def iter_fields(obj):
    for attribute_name in dir(obj):
        if attribute_name.startswith('_'):
            continue
        attribute = getattr(obj, attribute_name)
        if not isinstance(attribute, types.MethodType):
            yield attribute_name, attribute


def to_unicode(text):
    r"""
    >>> to_unicode('ascii') == u'ascii'
    True
    >>> to_unicode(u'utf\xe9'.encode('UTF-8')) == u'utf\xe9'
    True
    >>> to_unicode(u'unicode') == u'unicode'
    True
    """
    if isinstance(text, str):
        return text

    if isinstance(text, memoryview):
        text = text.tobytes()

    if not isinstance(text, bytes):
        return str(text)

    try:
        return text.decode('utf-8')
    except UnicodeError:
        pass
    try:
        return text.decode('iso-8859-15')
    except UnicodeError:
        pass
    return text.decode('windows-1252', 'replace')


def guess_encoding(stdio):
    try:
        encoding = stdio.encoding or locale.getpreferredencoding()
    except AttributeError:
        encoding = None
    # ASCII or ANSI is most likely a user mistake
    if not encoding or encoding.lower() == 'ascii' or encoding.lower().startswith('ansi'):
        encoding = 'UTF-8'
    return encoding


def limit(iterator, lim):
    """Iterate on the lim first elements of iterator."""
    count = 0
    iterator = iter(iterator)
    while count < lim:
        yield next(iterator)
        count += 1


def ratelimit(group, delay):
    """
    Simple rate limiting.

    Waits if the last call of lastlimit with this group name was less than
    delay seconds ago. The rate limiting is global, shared between any instance
    of the application and any call to this function sharing the same group
    name. The same group name should not be used with different delays.

    This function is intended to be called just before the code that should be
    rate-limited.

    This function is not thread-safe. For reasonably non-critical rate
    limiting (like accessing a website), it should be sufficient nevertheless.

    @param group [string]  rate limiting group name, alphanumeric
    @param delay [int]  delay in seconds between each call
    """

    from tempfile import gettempdir
    path = os.path.join(gettempdir(), 'woob_ratelimit.%s' % group)
    while True:
        try:
            offset = time() - os.stat(path).st_mtime
        except OSError:
            with open(path, 'w'):
                pass
            offset = 0

        if delay < offset:
            break

        sleep(delay - offset)

    os.utime(path, None)


def find_exe(basename):
    """
    Find the path to an executable by its base name (such as 'gpg').

    The executable can be overriden using an environment variable in the form
    `NAME_EXECUTABLE` where NAME is the specified base name in upper case.

    If the environment variable is not provided, the PATH will be searched
    both without and with a ".exe" suffix for Windows compatibility.

    If the executable can not be found, None is returned.
    """

    env_exe = os.getenv('%s_EXECUTABLE' % basename.upper())
    if env_exe and os.path.exists(env_exe) and os.access(env_exe, os.X_OK):
        return env_exe

    paths = os.getenv('PATH', os.defpath).split(os.pathsep)
    for path in paths:
        for ex in (basename, basename + '.exe'):
            fpath = os.path.join(path, ex)
            if os.path.exists(fpath) and os.access(fpath, os.X_OK):
                return fpath


def clean_text(
    text: str,
    *,
    remove_newlines: bool = True,
    normalize: Optional[Union[str, bool]] = 'NFC',
    transliterate: bool = False,
) -> str:
    """Clean a given text.

    This function is used by the
    :py:class:`woob.browser.filters.standard.CleanText` browser filter.

    :param text: The text to clean.
    :param remove_newlines: Whether to transform newlines into spaces or not.
        If this parameter is set to false, newlines will be normalized as
        UNIX-style newlines (U+000A).
    :param normalize: The Unicode normalization to apply, if relevant.
    :param transliterate: Whether to transliterate Unicode characters into
        ASCII characters, when possible.
    :return: The cleaned text.
    """
    # For compatibility with legacy code that sets ``normalize`` to a boolean,
    # we bind boolean values for the option to actual, usable values.
    if normalize is False:
        normalize = None
    elif normalize is True:
        normalize = 'NFC'

    if remove_newlines:
        text = NEWLINES_RE.sub(' ', text)
    else:
        # normalize newlines and clean what is inside
        text = '\n'.join(clean_text(line) for line in text.splitlines())

    text = text.strip()
    if normalize is not None:
        text = unicodedata.normalize(normalize, text)

    if transliterate:
        text = unidecode.unidecode(text)

    return text


def polling_loop(*, count=None, timeout=None, delay=5):
    """
    Delay iterator for polling loops.

    The count or timeout must be specified; both can be simultaneously.
    The default delay is five seconds.

    :param count: Maximum number of iterations this loop can produce.
                  Can be None for unlimited retries.
    :type count: int
    :param timeout: Maximum number of seconds to try the loop for.
    :type timeout: float
    :param delay: Delay in seconds between each iteration.
    :type delay: float
    """

    if count is None and timeout is None:
        raise ValueError(
            'polling_loop should be given at least a count or a timeout',
        )

    counter = itertools.count()
    if count is not None:
        counter = range(count)
    if timeout is not None:
        timeout = time() + timeout

    for idx in counter:
        if idx > 0:
            if timeout is not None and time() + delay >= timeout:
                break

            sleep(delay)

        yield idx


class classproperty:
    """
    Use it as a decorator to define a class property.

    >>> class C:
    ...     @classproperty
    ...     def VERSION(self):
    ...         return '3.3'
    ...
    >>> C.VERSION
    '3.3'
    >>> C().VERSION
    '3.3'
    """
    def __init__(self, f):
        self.f = f

    def __get__(self, obj, owner):
        return self.f(owner)



# Copyright(C) 2013 Julien Veyssier
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from woob.capabilities.subtitle import CapSubtitle
from woob.capabilities.base import empty
from woob.tools.application.repl import ReplApplication, defaultcount
from woob.tools.application.formatters.iformatter import IFormatter, PrettyFormatter


__all__ = ['AppSubtitles']

LANGUAGE_CONV = {
    'ar': 'ara', 'eo': 'epo',  'ga': '',    'ru': 'rus',
    'af': '', 'et': 'est',  'it': 'ita', 'sr': 'scc',
    'sq': 'alb', 'tl': '',  'ja': 'jpn', 'sk': 'slo',
    'hy': 'arm', 'fi': 'fin',  'kn': '',    'sl': 'slv',
    'az': '', 'fr': 'fre',  'ko': 'kor', 'es': 'spa',
    'eu': 'baq', 'gl': 'glg',  'la': '',    'sw': 'swa',
    'be': '', 'ka': 'geo',  'lv': 'lav', 'sv': 'swe',
    'bn': 'ben', 'de': 'ger',  'lt': 'lit', 'ta': '',
    'bg': 'bul', 'gr': 'ell',  'mk': 'mac', 'te': 'tel',
    'ca': 'cat', 'gu': '',  'ms': 'may', 'th': 'tha',
    'zh': 'chi', 'ht': '',  'mt': '',    'tr': 'tur',
    'hr': 'hrv', 'iw': 'heb',  'no': 'nor', 'uk': 'ukr',
    'cz': 'cze', 'hi': 'hin',  'fa': 'per', 'ur': 'urd',
    'da': 'dan', 'hu': 'hun',  'pl': 'pol', 'vi': 'vie',
    'nl': 'dut', 'is': 'ice',  'pt': 'por', 'cy': '',
    'en': 'eng', 'id': 'ind',  'ro': 'rum', 'yi': ''}


def sizeof_fmt(num):
    for x in ['bytes', 'KB', 'MB', 'GB', 'TB']:
        if num < 1024.0:
            return "%-4.1f%s" % (num, x)
        num /= 1024.0


class SubtitleInfoFormatter(IFormatter):
    MANDATORY_FIELDS = ('id', 'name', 'url', 'description')

    def format_obj(self, obj, alias):
        result = '%s%s%s\n' % (self.BOLD, obj.name, self.NC)
        result += 'ID: %s\n' % obj.fullid
        result += 'URL: %s\n' % obj.url
        if not empty(obj.language):
            result += 'LANG: %s\n' % obj.language
        if not empty(obj.nb_cd):
            result += 'NB CD: %s\n' % obj.nb_cd
        result += '\n%sDescription%s\n' % (self.BOLD, self.NC)
        result += '%s' % obj.description
        return result


class SubtitleListFormatter(PrettyFormatter):
    MANDATORY_FIELDS = ('id', 'name')

    def get_title(self, obj):
        return obj.name

    def get_description(self, obj):
        result = 'lang : %s' % obj.language
        result += ' ; %s CD' % obj.nb_cd
        if not empty(obj.url):
            result += ' ; url : %s' % obj.url
        return result


class AppSubtitles(ReplApplication):
    APPNAME = 'subtitles'
    VERSION = '3.7'
    COPYRIGHT = 'Copyright(C) 2013-YEAR Julien Veyssier'
    DESCRIPTION = "Console application allowing to search for subtitles on various services " \
                  "and download them."
    SHORT_DESCRIPTION = "search and download subtitles"
    CAPS = CapSubtitle
    EXTRA_FORMATTERS = {'subtitle_list': SubtitleListFormatter,
                        'subtitle_info': SubtitleInfoFormatter
                        }
    COMMANDS_FORMATTERS = {'search':    'subtitle_list',
                           'info':      'subtitle_info'
                           }

    def complete_info(self, text, line, *ignored):
        args = line.split(' ')
        if len(args) == 2:
            return self._complete_object()

    def do_info(self, id):
        """
        info ID

        Get information about a subtitle.
        """

        subtitle = self.get_object(id, 'get_subtitle')
        if not subtitle:
            print('Subtitle not found: %s' % id, file=self.stderr)
            return 3

        self.start_format()
        self.format(subtitle)

    def complete_download(self, text, line, *ignored):
        args = line.split(' ', 2)
        if len(args) == 2:
            return self._complete_object()
        elif len(args) >= 3:
            return self.path_completer(args[2])

    def do_download(self, line):
        """
        download ID [FILENAME]

        Get the subtitle or archive file.
        FILENAME is where to write the file. If FILENAME is '-',
        the file is written to stdout.
        """
        id, dest = self.parse_command_args(line, 2, 1)

        subtitle = self.get_object(id, 'get_subtitle')
        if not subtitle:
            print('Subtitle not found: %s' % id, file=self.stderr)
            return 3

        if dest is None:
            ext = subtitle.ext
            if empty(ext):
                ext = 'zip'
            dest = '%s.%s' % (subtitle.name, ext)

        for buf in self.do('get_subtitle_file', subtitle.id, backends=subtitle.backend):
            if buf:
                if dest == '-':
                    self.stdout.buffer.write(buf)
                else:
                    try:
                        with open(dest, 'wb') as f:
                            f.write(buf)
                    except IOError as e:
                        print('Unable to write file in "%s": %s' % (dest, e), file=self.stderr)
                        return 1
                    else:
                        print('Saved to %s' % dest)
                return

    @defaultcount(10)
    def do_search(self, line):
        """
        search language [PATTERN]

        Search subtitles.

        Language  Abbreviation
        ----------------------
        Arabic      ar          Esperanto   eo          Irish       ga          Russian     ru
        Afrikaans   af          Estonian    et          Italian     it          Serbian     sr
        Albanian    sq          Filipino    tl          Japanese    ja          Slovak      sk
        Armenian    hy          Finnish     fi          Kannada     kn          Slovenian   sl
        Azerbaijani az          French      fr          Korean      ko          Spanish     es
        Basque      eu          Galician    gl          Latin       la          Swahili     sw
        Belarusian  be          Georgian    ka          Latvian     lv          Swedish     sv
        Bengali     bn          German      de          Lithuanian  lt          Tamil       ta
        Bulgarian   bg          Greek       gr          Macedonian  mk          Telugu      te
        Catalan     ca          Gujarati    gu          Malay       ms          Thai        th
        Chinese     zh          Haitian     ht          Maltese     mt          Turkish     tr
        Croatian    hr          Hebrew      iw          Norwegian   no          Ukrainian   uk
        Czech       cz          Hindi       hi          Persian     fa          Urdu        ur
        Danish      da          Hungaric    hu          Polish      pl          Vietnamese  vi
        Dutch       nl          Icelandic   is          Portuguese  pt          Welsh       cy
        English     en          Indonesian  id          Romanian    ro          Yiddish     yi
        ----------------------
        """
        language, pattern = self.parse_command_args(line, 2, 1)
        self.change_path(['search'])
        if not pattern:
            pattern = None

        self.start_format(pattern=pattern)
        for subtitle in self.do('iter_subtitles', language=language, pattern=pattern):
            self.cached_format(subtitle)



# Copyright(C) 2023 Powens
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

# flake8: compatible

"""Utilities for handling PKCE (RFC 7636)."""

from __future__ import annotations

from base64 import urlsafe_b64encode
from enum import Enum
from hashlib import sha256
from os import urandom
from typing import NamedTuple

__all__ = [
    'PKCEChallengeType',
    'PKCEData',
]


class PKCEChallengeType(str, Enum):
    """PKCE challenge type."""

    PLAIN = ('PLAIN')
    """Plaintext challenge."""

    S256 = ('S256')
    """SHA-256 challenge."""


class PKCEData(NamedTuple):
    """PKCE data to generate."""

    verifier: str
    """The verifier to transmit at token call."""

    challenge: str
    """The challenge to transmit in the authorization URL."""

    method: str
    """The method to transmit in the authorization URL."""

    @classmethod
    def build(
        cls: type[PKCEData],
        type_: PKCEChallengeType = PKCEChallengeType.S256,
    ) -> PKCEData:
        r"""Build random data for OAuth2 PKCE extension.

        :param type\_: The type of challenge to produce.
        :return: The PKCE data.
        """
        # While RFC 7636 section 4.1 allows verifiers to use the whole uppercase and
        # lowercase latin alphabet, some APIs restrict the verifier charset to hex
        # characters (0-9, a-f), hence the limitation employed here.
        #
        # RFC 7636 section 7.1 mandates 256 bits of entropy (32 8-bit bytes), and
        # section 4.1 mandates between 43 and 128 characters in length; this method
        # makes us generate 64 characters with enough entropy, we're in the clear!
        verifier = urandom(32).hex()

        if type_ == PKCEChallengeType.S256:
            digest = sha256(verifier.encode('ascii')).digest()
            challenge = urlsafe_b64encode(digest).rstrip(b'=').decode('ascii')
            return cls(verifier=verifier, method='S256', challenge=challenge)

        if type_ == PKCEChallengeType.PLAIN:
            return cls(verifier=verifier, method='plain', challenge=verifier)

        raise ValueError(f'Unsupported PKCE challenge type "{type_}"')



# Copyright(C) 2010-2023 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import logging
import importlib
import os
from copy import copy
from threading import RLock
import warnings
from urllib.request import getproxies
from typing import List, Any, Dict, Tuple, Iterator, Type, ClassVar, TYPE_CHECKING

from packaging.version import Version

from woob import __version__
from woob.capabilities.base import BaseObject, Capability, FieldNotFound, NotAvailable, NotLoaded
from woob.tools.json import json
from woob.tools.log import getLogger
from woob.tools.misc import iter_fields
from woob.tools.storage import IStorage
from woob.tools.value import ValuesDict, ValueBool

if TYPE_CHECKING:
    from woob.core import WoobBase
    from woob.browser import Browser


__all__ = ['BackendStorage', 'BackendConfig', 'Module']


class BackendStorage:
    """
    This is an abstract layer to store data in storages (:mod:`woob.tools.storage`)
    easily.

    It is instancied automatically in constructor of :class:`Module`, in the
    :attr:`Module.storage` attribute.

    :param name: name of backend
    :type name: str
    :param storage: storage object
    :type storage: :class:`woob.tools.storage.IStorage`
    """

    def __init__(self, name: str, storage: IStorage | None):
        self.name = name
        self.storage = storage

    def set(self, *args):
        """
        Set value in the storage.

        Example:

        >>> from woob.tools.storage import StandardStorage
        >>> backend = BackendStorage('blah', StandardStorage('/tmp/cfg'))
        >>> backend.storage.set('config', 'nb_of_threads', 10)  # doctest: +SKIP
        >>>

        :param args: the path where to store value
        """
        if self.storage:
            self.storage.set('backends', self.name, *args)

    def delete(self, *args):
        """
        Delete a value from the storage.

        :param args: path to delete.
        """
        if self.storage:
            self.storage.delete('backends', self.name, *args)

    def get(self, *args, **kwargs) -> Any:
        """
        Get a value or a dict of values in storage.

        Example:

        >>> from woob.tools.storage import StandardStorage
        >>> backend = BackendStorage('blah', StandardStorage('/tmp/cfg'))
        >>> backend.storage.get('config', 'nb_of_threads')  # doctest: +SKIP
        10
        >>> backend.storage.get('config', 'unexistant', 'path', default='lol')  # doctest: +SKIP
        'lol'
        >>> backend.storage.get('config')  # doctest: +SKIP
        {'nb_of_threads': 10, 'other_things': 'blah'}

        :param args: path to get
        :param default: if specified, default value when path is not found
        """
        if self.storage:
            return self.storage.get('backends', self.name, *args, **kwargs)

        return kwargs.get('default', None)

    def load(self, default: Dict):
        """
        Load storage.

        It is made automatically when your backend is created, and use the
        ``STORAGE`` class attribute as default.

        :param default: this is the default tree if storage is empty
        :type default: :class:`dict`
        """
        if self.storage:
            self.storage.load('backends', self.name, default)

    def save(self):
        """
        Save storage.
        """
        if self.storage:
            self.storage.save('backends', self.name)


class BackendConfig(ValuesDict):
    """
    Configuration of a backend.

    This class is firstly instanced as a :class:`woob.tools.value.ValuesDict`,
    containing some :class:`woob.tools.value.Value` (and derivated) objects.

    Then, using the :func:`load` method will load configuration from file and
    create a copy of the :class:`BackendConfig` object with the loaded values.
    """
    modname: str
    instname: str
    woob: WoobBase

    def load(
        self,
        woob: WoobBase,
        modname: str,
        instname: str,
        config: Dict,
        nofail: bool = False
    ) -> BackendConfig:
        """
        Load configuration from dict to create an instance.

        :param woob: woob object
        :type woob: :class:`woob.core.woob.WoobBase`
        :param modname: name of the module
        :type modname: :class:`str`
        :param instname: name of this backend
        :type instname: :class:`str`
        :param params: parameters to load
        :type params: :class:`dict`
        :param nofail: if true, this call can't fail
        :type nofail: :class:`bool`
        :rtype: :class:`BackendConfig`
        """
        cfg = self.__class__()
        cfg.modname = modname
        cfg.instname = instname
        cfg.woob = woob
        for name, field in self.items():
            value = config.get(name, None)

            if value is None:
                if not nofail and field.required:
                    raise Module.ConfigError(
                        f'Backend({cfg.instname}): Configuration error: Missing parameter {name} ({field.description})',
                       bad_fields=[name]
                    )
                value = field.default

            field = copy(field)
            try:
                field.load(cfg.instname, value, cfg.woob.requests)
            except ValueError as v:
                if not nofail:
                    raise Module.ConfigError(
                        f'Backend({cfg.instname}): Configuration error for field "{name}": {v}',
                        bad_fields=[name]
                    )

            cfg[name] = field
        return cfg

    def dump(self) -> dict:
        """
        Dump config in a dictionary.

        :rtype: :class:`dict`
        """
        settings = {}
        for name, value in self.items():
            if not value.transient:
                settings[name] = value.dump()
        return settings

    def save(self, edit: bool = True, params: Dict | None = None):
        """
        Save backend config.

        :param edit: if true, it changes config of an existing backend
        :type edit: :class:`bool`
        :param params: if supplied, params to merge with the ones of the current object
        :type params: :class:`dict`
        """
        assert self.modname is not None
        assert self.instname is not None
        assert self.woob is not None

        dump = self.dump()
        if params is not None:
            dump.update(params)

        if edit:
            self.woob.backends_config.edit_backend(self.instname, dump)
        else:
            self.woob.backends_config.add_backend(self.instname, self.modname, dump)


class Module:
    """
    Base class for modules.

    You may derivate it, and also all capabilities you want to implement.

    :param woob: woob instance
    :type woob: :class:`woob.core.woob.Woob`
    :param name: name of backend
    :type name: :class:`str`
    :param config: configuration of backend (optional)
    :type config: :class:`dict`
    :param storage: storage object (optional)
    :type storage: :class:`woob.tools.storage.IStorage`
    :param logger: parent logger (optional)
    :type logger: :class:`logging.Logger`
    """

    NAME: ClassVar[str]
    """Name of the maintainer of this module."""

    MAINTAINER: ClassVar[str] = '<unspecified>'
    """Name of the maintainer."""

    EMAIL: ClassVar[str] = '<unspecified>'
    """Email address of the maintainer."""

    DESCRIPTION: ClassVar[str] = '<unspecified>'
    """Description"""

    LICENSE: ClassVar[str] = '<unspecified>'
    """License of the module"""

    CONFIG: ClassVar[BackendConfig] = BackendConfig()
    """Configuration required for backends.

    Values must be :class:`woob.tools.value.Value` objects.
    """

    STORAGE: ClassVar[Dict] = {}
    """Storage"""

    BROWSER: Browser | None = None
    """Browser class"""

    ICON: ClassVar[str | None] = None
    """URL to an optional icon.

    If you want to create your own icon, create a 'favicon.png' icon in
    the module's directory, and keep the ICON value to None.
    """

    OBJECTS: ClassVar[Dict] = {}
    """Supported objects to fill

    The key is the class and the value the method to call to fill
    Method prototype: method(object, fields)
    When the method is called, fields are only the one which are
    NOT yet filled.
    """

    DEPENDENCIES: ClassVar[Tuple[str]] = ()
    """Tuple of module names on which this module depends."""

    class ConfigError(Exception):
        """
        Raised when the config can't be loaded.
        """

        def __init__(self, message, bad_fields=None):
            """
            :type message: str
            :param message: message of the exception
            :type bad_fields: list[str]
            :param bad_fields: names of the config fields which are incorrect
            """

            super().__init__(message)
            self.bad_fields = bad_fields or ()

    def __enter__(self):
        self.lock.acquire()

    def __exit__(self, t, v, tb):
        self.lock.release()

    def __repr__(self):
        return f"<Backend {self.name}>"

    def __new__(cls, *args, **kwargs):
        """ Accept any arguments, necessary for AbstractModule __new__ override.

        AbstractModule, in its overridden __new__, removes itself from class hierarchy
        so its __new__ is called only once. In python 3, default (object) __new__ is
        then used for next instantiations but it's a slot/"fixed" version supporting
        only one argument (type to instanciate).
        """
        return object.__new__(cls)

    @property
    def VERSION(self):
        warnings.warn(
            'Attribute Module.VERSION will be removed in woob 4, do not use it.',
            DeprecationWarning,
            stacklevel=3
        )
        return Version(__version__).base_version

    def __init__(
        self,
        woob: WoobBase,
        name: str,
        config: Dict | None = None,
        storage: IStorage | None = None,
        logger: logging.Logger | None = None,
        nofail: bool = False
    ):
        if (
            hasattr(self.__class__, 'VERSION') and
            not isinstance(self.__class__.VERSION, property)
        ):
            warnings.warn(
                f'Class attribute {self.__class__.__name__}.VERSION is now '
                'unused and deprecated, you can remove it. '
                'If you do so, do not forget to increase the woob version to at '
                'least 3.4 in requirements.txt.',
                DeprecationWarning,
            )

        self.logger = getLogger(name, parent=logger)
        self.woob = woob
        self.name = name
        self.lock = RLock()
        if config is None:
            config = {}

        # Private fields (which start with '_')
        self._private_config = dict((key, value) for key, value in config.items() if key.startswith('_'))

        # Load configuration of backend.
        self.config = self.CONFIG.load(woob, self.NAME, self.name, config, nofail)

        self.storage = BackendStorage(self.name, storage)
        self.storage.load(self.STORAGE)

    def dump_state(self):
        """
        Dump module state into storage.
        """
        if hasattr(self.browser, 'dump_state'):
            self.storage.set('browser_state', self.browser.dump_state())
            self.storage.save()

    def deinit(self):
        """
        This abstract method is called when the backend is unloaded.
        """
        if self._browser is None:
            return

        try:
            self.dump_state()
        finally:
            if hasattr(self.browser, 'deinit'):
                self.browser.deinit()

    @property
    def weboob(self):
        """
        .. deprecated:: 3.4
           Don't use this attribute, but :attr:`woob` instead.
        """
        warnings.warn('Use Module.woob instead.', DeprecationWarning, stacklevel=2)
        return self.woob

    _browser = None

    @property
    def browser(self) -> Browser:
        """
        Attribute 'browser'. The browser is created at the first call
        of this attribute, to avoid useless pages access.

        Note that the :func:`create_default_browser` method is called to create it.
        """
        if self._browser is None:
            self._browser = self.create_default_browser()
        return self._browser

    def create_default_browser(self) -> Browser | None:
        """
        Method to overload to build the default browser in
        attribute 'browser'.
        """
        return self.create_browser()

    def create_browser(self, *args, **kwargs) -> Browser | None:
        """
        Build a browser from the BROWSER class attribute and the
        given arguments.

        :param klass: optional parameter to give another browser class to instanciate
        :type klass: :class:`woob.browser.browsers.Browser`
        :param load_state: Whether to load the browser state if it supports
            the feature.
        :type load_state: bool
        """

        klass = kwargs.pop('klass', self.BROWSER)
        if not klass:
            return None

        should_load_state = bool(kwargs.pop('load_state', True))

        kwargs['proxy'] = self.get_proxy()
        if '_proxy_headers' in self._private_config:
            kwargs['proxy_headers'] = self._private_config['_proxy_headers']
            if isinstance(kwargs['proxy_headers'], str):
                kwargs['proxy_headers'] = json.loads(kwargs['proxy_headers'])

        if '_ssl_verify' in self._private_config:
            # value can be either a boolean or a string (path)
            value = ValueBool()
            try:
                value.set(self._private_config['_ssl_verify'])
            except ValueError:
                kwargs.setdefault('verify', self._private_config['_ssl_verify'])
            else:
                kwargs.setdefault('verify', value.get())

        kwargs['logger'] = self.logger

        if self.logger.settings['responses_dirname']:
            kwargs.setdefault('responses_dirname', os.path.join(self.logger.settings['responses_dirname'],
                                                                self._private_config.get('_debug_dir', self.name)))
        elif os.path.isabs(self._private_config.get('_debug_dir', '')):
            kwargs.setdefault('responses_dirname', self._private_config['_debug_dir'])

        if '_highlight_el' in self._private_config:
            value = ValueBool()
            try:
                value.set(self._private_config['_highlight_el'])
            except ValueError as e:
                raise Module.ConfigError(
                    f'Backend({self.name}): Configuration error: _highlight_el must be a boolean'
                ) from e

            kwargs.setdefault('highlight_el', value.get())

        browser = klass(*args, **kwargs)

        if should_load_state and hasattr(browser, 'load_state'):
            browser.load_state(self.storage.get('browser_state', default={}))

        return browser

    def get_proxy(self) -> Dict[str, str]:
        """
        Get proxy to use.

        It will read in environment variables, then in backend config.

        Proxy keys in backend config are:

        * ``_proxy`` for HTTP requests
        * ``_proxy_ssl`` for HTTPS requests
        """
        # Get proxies from environment variables
        proxies = getproxies()
        # Override them with backend-specific config
        if '_proxy' in self._private_config:
            proxies['http'] = self._private_config['_proxy']
        if '_proxy_ssl' in self._private_config:
            proxies['https'] = self._private_config['_proxy_ssl']
        # Remove empty values
        for key in list(proxies.keys()):
            if not proxies[key]:
                del proxies[key]
        return proxies

    @classmethod
    def iter_caps(cls) -> Iterator[Type[Capability]]:
        """
        Iter capabilities implemented by this backend.

        :rtype: iter[:class:`woob.capabilities.base.Capability`]
        """
        for base in cls.mro():
            if issubclass(base, Capability) and base != Capability and base != cls and not issubclass(base, Module):
                yield base

    def has_caps(self, *caps) -> bool:
        """
        Check if this backend implements at least one of these capabilities.

        `caps` should be list of :class:`Capability` objects (e.g. :class:`CapBank`) or capability names (e.g. 'bank').
        """
        available_cap_names = [cap.__name__ for cap in self.iter_caps()]
        return any(
            (isinstance(c, str) and c in available_cap_names) or isinstance(self, c)
            for c in caps
        )

    def fillobj(self, obj: BaseObject, fields: List[str] | None = None):
        """
        Fill an object with the wanted fields.

        :param fields: what fields to fill; if None, all fields are filled
        :type fields: :class:`list`
        """
        if obj is None:
            return obj

        def not_loaded_or_incomplete(v):
            return (v is NotLoaded or isinstance(v, BaseObject) and not v.__iscomplete__())

        def not_loaded(v):
            return v is NotLoaded

        def filter_missing_fields(obj, fields, check_cb):
            missing_fields = []
            if fields is None:
                # Select all fields
                if isinstance(obj, BaseObject):
                    fields = [item[0] for item in obj.iter_fields()]
                else:
                    fields = [item[0] for item in iter_fields(obj)]

            for field in fields:
                if not hasattr(obj, field):
                    raise FieldNotFound(obj, field)
                value = getattr(obj, field)

                missing = False
                if hasattr(value, '__iter__'):
                    for v in (value.values() if isinstance(value, dict) else value):
                        if check_cb(v):
                            missing = True
                            break
                elif check_cb(value):
                    missing = True

                if missing:
                    missing_fields.append(field)

            return missing_fields

        if isinstance(fields, str):
            fields = (fields,)

        missing_fields = filter_missing_fields(obj, fields, not_loaded_or_incomplete)

        if not missing_fields:
            return obj

        for key, value in self.OBJECTS.items():
            if isinstance(obj, key):
                self.logger.debug('Fill %r with fields: %s', obj, missing_fields)
                obj = value(self, obj, missing_fields) or obj
                break

        missing_fields = filter_missing_fields(obj, fields, not_loaded)

        # Object is not supported by backend. Do not notice it to avoid flooding user.
        # That's not so bad.
        for field in missing_fields:
            setattr(obj, field, NotAvailable)

        return obj


class AbstractModuleMissingParentError(Exception):
    pass


class MetaModule(type):
    # we can remove this class as soon as we get rid of Abstract*
    def __new__(mcs, name, bases, dct):
        if name != 'AbstractModule' and AbstractModule in bases:
            warnings.warn('AbstractModule is deprecated and will be removed in woob 4.0. '
                          'Use standard "from woob_modules.other_module import Module" instead.',
                          DeprecationWarning, stacklevel=2)

            module = importlib.import_module('woob_modules.%s' % dct['PARENT'])
            symbols = [getattr(module, attr) for attr in dir(module) if not attr.startswith('__')]
            klass = next(
                symbol for symbol in symbols
                if isinstance(symbol, type) and issubclass(symbol, Module) and symbol != Module
            )

            bases = tuple(klass if isinstance(base, mcs) else base for base in bases)

            additional_config = dct.pop('ADDITIONAL_CONFIG', None)
            if additional_config:
                dct['CONFIG'] = BackendConfig(*(list(klass.CONFIG.values()) + list(additional_config.values())))

        return super().__new__(mcs, name, bases, dct)


class AbstractModule(metaclass=MetaModule):
    """
    .. deprecated:: 3.4
       Don't use this class, import woob_modules.other_module.etc instead
    """



# Copyright(C) 2010-2011 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.


class ConfigError(Exception):
    pass


class IConfig:
    """
    Interface for config storage.

    Config stores keys and values. Each key is a path of components, allowing
    to group multiple options.
    """

    def load(self, default={}):
        """
        Load config.

        :param default: default values for the config
        :type default: dict[:class:`str`]
        """
        raise NotImplementedError()

    def save(self):
        """Save config."""
        raise NotImplementedError()

    def set(self, *args):
        """
        Set a config value.

        :param args: all args except the last arg are the path of the option key.
        :type args: str or object
        """
        raise NotImplementedError()

    def delete(self, *args):
        """
        Delete an option from config.

        :param args: path to the option key.
        :type args: str
        """
        raise NotImplementedError()

    def get(self, *args, **kwargs):
        """
        Get the value of an option.

        :param args: path of the option key.
        :param default: if specified, default value when path is not found
        """
        raise NotImplementedError()

    def __enter__(self):
        self.load()

    def __exit__(self, t, v, tb):
        self.save()



# Copyright(C) 2010-2021  Nicolas Duhamel
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from collections import OrderedDict

from .base import Capability, BaseObject, UserError, StringField, Field


__all__ = ['CapCollection', 'BaseCollection', 'Collection', 'CollectionNotFound']


class CollectionNotFound(UserError):
    def __init__(self, split_path=None):
        if split_path is not None:
            msg = 'Collection not found: %s' % '/'.join(split_path)
        else:
            msg = 'Collection not found'
        super(CollectionNotFound, self).__init__(msg)


class BaseCollection(BaseObject):
    """
    Inherit from this if you want to create an object that is *also* a Collection.
    However, this probably will not work properly for now.
    """

    def __init__(self, split_path, id=None, url=None):
        super(BaseCollection, self).__init__(id, url)
        self.split_path = split_path

    @property
    def basename(self):
        return self.split_path[-1] if self.path_level else None

    @property
    def parent_path(self):
        return self.split_path[0:-1] if self.path_level else None

    @property
    def path_level(self):
        return len(self.split_path)

    def to_dict(self):
        def iter_decorate(d):
            for key, value in d:
                if key == 'id' and self.backend is not None:
                    value = '%s@%s' % (self.basename, self.backend)
                yield key, value

                if key == 'split_path':
                    yield key, '/'.join(value)

        fields_iterator = self.iter_fields()
        return OrderedDict(iter_decorate(fields_iterator))


class Collection(BaseCollection):
    """
    A Collection is a "fake" object returned in results, which shows you can get
    more results if you go into its path.

    It is a dumb object, it must not contain callbacks to a backend.

    Do not inherit from this class if you want to make a regular BaseObject
    a Collection, use BaseCollection instead.
    """
    title = StringField('Collection title')
    split_path = Field('Full collection path', list)

    def __init__(self, split_path=None, title=None, id=None, url=None):
        self.title = title
        super(Collection, self).__init__(split_path, id, url)

    def __str__(self):
        if self.title and self.basename:
            return '%s (%s)' % (self.basename, self.title)
        elif self.basename:
            return '%s' % self.basename
        else:
            return 'Unknown collection'


class CapCollection(Capability):
    def iter_resources_flat(self, objs, split_path, clean_only=False):
        """
        Call iter_resources() to fetch all resources in the tree.
        If clean_only is True, do not explore paths, only remove them.
        split_path is used to set the starting path.
        """
        for resource in self.iter_resources(objs, split_path):
            if isinstance(resource, Collection):
                if not clean_only:
                    for res in self.iter_resources_flat(objs, resource.split_path):
                        yield res
            else:
                yield resource

    def iter_resources(self, objs, split_path):
        """
        split_path is a list, either empty (root path) or with one or many
        components.
        """
        raise NotImplementedError()

    def get_collection(self, objs, split_path):
        """
        Get a collection for a given split path.
        If the path is invalid (i.e. can't be handled by this module),
        it should return None.
        """
        collection = Collection(split_path, None)
        return self.validate_collection(objs, collection) or collection

    def validate_collection(self, objs, collection):
        """
        Tests if a collection is valid.
        For compatibility reasons, and to provide a default way, it checks if
        the collection has at least one object in it. However, it is not very
        efficient or exact, and you are encouraged to override this method.
        You can replace the collection object entirely by returning a new one.
        """
        # Root
        if collection.path_level == 0:
            return
        try:
            i = self.iter_resources(objs, collection.split_path)
            next(i)
        except StopIteration:
            raise CollectionNotFound(collection.split_path)

    def _restrict_level(self, split_path, lmax=0):
        if len(split_path) > lmax:
            raise CollectionNotFound(split_path)



# Copyright(C) 2010-2014 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

from typing import Dict

from woob.tools.value import Value

from .base import Capability, BaseObject, Field, StringField, UserError
from .date import DateField
from .contact import Contact



__all__ = ['OptimizationNotFound', 'Optimization', 'Event', 'CapDating']


class OptimizationNotFound(UserError):
    """
    Raised when an optimization is not found.
    """


class Optimization(BaseObject):
    """
    Optimization.

    :var CONFIG: Configuration of optim can be made by
                 :class:`woob.tools.value.Value` objects
                 in this dict.
    """
    CONFIG: Dict[str, Value] = {}

    def start(self):
        """
        Start optimization.
        """
        raise NotImplementedError()

    def stop(self):
        """
        Stop optimization.
        """
        raise NotImplementedError()

    def is_running(self):
        """
        Know if the optimization is currently running.

        :rtype: bool
        """
        raise NotImplementedError()

    def get_config(self):
        """
        Get config of this optimization.

        :rtype: dict
        """
        return None

    def set_config(self, params):
        """
        Set config of this optimization.

        :param params: parameters
        :type params: dict
        """
        raise NotImplementedError()


class Event(BaseObject):
    """
    A dating event (for example a visite, a query received, etc.)
    """
    date =      DateField('Date of event')
    contact =   Field('Contact related to this event', Contact)
    type =      StringField('Type of event')
    message =   StringField('Message of the event')


class CapDating(Capability):
    """
    Capability for dating websites.
    """

    def init_optimizations(self):
        """
        Initialization of optimizations.
        """
        raise NotImplementedError()

    def add_optimization(self, name, optim):
        """
        Add an optimization.

        :param name: name of optimization
        :type name: str
        :param optim: optimization
        :type optim: :class:`Optimization`
        """
        optim.id = name
        setattr(self, 'OPTIM_%s' % name, optim)

    def iter_optimizations(self):
        """
        Iter optimizations.

        :rtype: iter[:class:`Optimization`]
        """
        for attr_name in dir(self):
            if not attr_name.startswith('OPTIM_'):
                continue
            attr = getattr(self, attr_name)
            if attr is None:
                continue

            yield attr

    def get_optimization(self, optim):
        """
        Get an optimization from a name.

        :param optim: name of optimization
        :type optim: str
        :rtype: :class:`Optimization`
        """
        optim = optim.upper()
        if not hasattr(self, 'OPTIM_%s' % optim):
            raise OptimizationNotFound()

        return getattr(self, 'OPTIM_%s' % optim)

    def iter_events(self):
        """
        Iter events.

        :rtype: iter[:class:`Event`]
        """
        raise NotImplementedError()

    def iter_new_contacts(self):
        """
        Iter new contacts.

        :rtype: iter[:class:`Contact`]
        """
        raise NotImplementedError()



# Copyright(C) 2019-2020 Célande Adrien
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.


from woob.capabilities.base import find_object

from .base import (
    BaseObject,
    Field, StringField,  BoolField,
    EnumField, Enum,
    UserError,
)
from .collection import CapCollection

__all__ = ['CapRPG', 'Character', 'Skill', 'CharacterClass', 'CollectableItem']


class CharacterNotFound(UserError):
    """ Raised when a character is not found """
    def __init__(self, msg='Character not found'):
        super(CharacterNotFound, self).__init__(msg)


class SkillNotFound(UserError):
    """ Raised when a skill is not found """
    def __init__(self, msg='Skill not found'):
        super(SkillNotFound, self).__init__(msg)


class CharacterClassNotFound(UserError):
    """ Raised when a class is not found """
    def __init__(self, msg='Class not found'):
        super(CharacterClassNotFound, self).__init__(msg)


class CollectableItemNotFound(UserError):
    """ Raised when an item is not found """
    def __init__(self, msg='Item not found'):
        super(CollectableItemNotFound, self).__init__(msg)


class ListField(Field):
    """ Field made of a list """
    def __init__(self, doc, **kwargs):
        if 'default' not in kwargs:
            kwargs['default'] = []
        super(ListField, self).__init__(doc, list, **kwargs)


class DictField(Field):
    """ Field made of a dict """
    def __init__(self, doc, **kwargs):
        if 'default' not in kwargs:
            kwargs['default'] = {}
        super(DictField, self).__init__(doc, dict, **kwargs)


class BaseRPGObject(BaseObject):
    """ Object used to build up all the objects of the CapRPG """
    # TODO: check id and origin for duplicates
    name = StringField('Name', mandatory=True)
    description = StringField('Description', mandatory=False)
    origin = StringField('From which game/platform the object comes from', mandatory=False)
    picture = StringField('URL of a picture', mandatory=False)


class CharacterClass(BaseRPGObject):
    """ CharacterClass of a character """
    pass


class SkillType(Enum):
    """
    Types of skill
    """
    UNKNOWN = 0
    """UNKNOWN: Unknown type"""

    PASSIVE = 1
    """PASSIVE: The effects of the skill are always on. The skill does not need to be activated."""

    ACTIVE = 2
    """ACTIVE: The skill must be activated."""


class SkillTarget(Enum):
    """
    Target of a skill
    """
    UNKNOWN = 0
    """UNKNOWN: Unknown target"""

    SELF = 1
    """SELF: The target is the one that launched it"""

    FOE = 2
    """FOE: The target is a different character than the one who launched it"""

    SELF_AND_FOE = 3
    """SELF_AND_FOE: The target is a different character and the one who lanched it"""

    FIELD = 4
    """FIELD: The target is the field or the environment"""


class SkillCategory(Enum):
    """
    Category of a skill
    """
    UNKNOWN = 0
    """UNKNOWN: Unknown category"""

    PHYSICAL = 1
    """PHYSICAL: The skill is meant to deal physical damages (and effects)"""

    MAGICAL = 2
    """MAGICAL: The skill is meant to deal magical or special damages (and effects)"""

    PHYSICAL_AND_MAGICAL = 3
    """PHYSICAL_AND_MAGICAM: The skill is meant to deal physical and magical damages (and effects)"""

    STATUS = 4
    """STATUS: The skill does not deal direct damages."""


class Skill(BaseRPGObject):
    """ Skill of a character """
    type = EnumField('Type of skill', SkillType, mandatory=True, default=SkillType.UNKNOWN)
    target = EnumField('Target of the skill', SkillTarget, mandatory=True, default=SkillTarget.UNKNOWN)

    statistics = DictField('Dict of statistics', mandatory=False)
    character_classes = ListField('List of CharacterClass ids that can use this move', mandatory=False)
    category = EnumField('Category of skill', SkillCategory, mandatory=False)


class Character(BaseRPGObject):
    """ Creature or person """
    base_stats = DictField('Base statistics')

    character_classes = ListField('List of CharacterClasses id', mandatory=False)
    skills = ListField('List of Skills id', mandatory=False)
    next_forms = ListField('List of the next forms of the character', mandatory=False)
    locations = ListField('List of locations of the character', mandatory=False)


class CollectableItem(BaseRPGObject):
    """ Object that you can find in the game """
    to_use = BoolField('The object can be used at anytime', mandatory=False)
    to_carry = BoolField('The object must be carried to be used (like in battle)', mandatory=False)
    category = StringField('Category of the item', mandatory=False)
    locations = ListField('List of locations of the item', mandatory=False)


class CapRPG(CapCollection):
    """
    Capability for rpg games to list characters, objects, etc.
    """
    def iter_resources(self, objs, split_path):
        """
        Iter reources.

        return :func:`iter_characters` for 'character'
        """
        if Character in objs:
            self._restrict_level(split_path)
            return self.iter_characters()


    def iter_characters(self):
        """
        Iter characters.

        :rtype: iter[:class: `Character`]
        """
        raise NotImplementedError()

    def get_character(self, character_id):
        """
        Get a character with its ID.

        :param character_id: ID of the character
        :type character_id: :class:`str`
        :rtype: :class: `Character`
        :raises: :class: `CharcterNotFound`
        """
        # TODO: find from id and version?
        return find_object(self.iter_characters(), id=character_id, error=CharacterNotFound)

    def iter_skills(self, skill_type=None):
        """
        Iter all available skills.

        :param skill_type: Type of skill
        :type skill_type: :class:`int`
        :rtype: iter[:class: `Skill`]
        """
        raise NotImplementedError()

    def get_skill(self, skill_id):
        """
        Get a skill from with ID.

        :param skill_id: ID of the skill
        :type skill_id: :class:`str`
        :rtype: :class: `Skill`
        :raises: :class: `SkillNotFound`
        """
        return find_object(self.iter_skills(), id=skill_id, error=SkillNotFound)

    def iter_skill_set(self, character_id, skill_type=None):
        """
        Iter skills for a specific character

        :param character_id: ID of the character
        :type character_id: :class:`str`
        :param skill_type: Type of skill
        :type skill_type: :class:`int`
        :rtype: :class: iter[:class: `Skill`]
        """
        raise NotImplementedError()

    def iter_character_classes(self):
        """
        Iter all classes

        :rtype: :class: iter[:class: `CharacterClass`]
        """
        raise NotImplementedError()

    def get_character_class(self, class_id):
        """
        Get details of a class according to id

        :param class_id: ID of the skill
        :type class_id: :class:`str`
        :rtype: :class: `CharacterClass`
        :raises: :class: `CharacterClassNotFound`
        """
        return find_object(self.iter_character_classes(), id=class_id, error=CharacterClassNotFound)

    def iter_collectable_items(self):
        """
        Iter all collectable items

        :rtype: :class: iter[:class: `CollectableItem`]
        """
        raise NotImplementedError()

    def get_collectable_item(self, item_id):
        """
        Get details of a collectable item according to id

        :param item_id: ID of the skill
        :type item_id: :class:`str`
        :rtype: :class: `CollectableItem`
        :raises: :class: `CollectableItemNotFound`
        """
        return find_object(self.iter_collectable_items(), id=item_id, error=CollectableItemNotFound)



# -*- coding: utf-8 -*-

# Copyright(C) 2010-2021 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

import sys
from importlib.abc import MetaPathFinder, Loader
import importlib

import woob


__title__ = woob.__title__
__version__ = woob.__version__
__author__ = woob.__author__
__copyright__ = woob.__copyright__


# Use case: the core does "import woob.x" but some app does "import weboob.x".
# If we merely set __path__, python will generate different modules
# though they have the same path. So the references will be different,
# isinstance(woob.X(), weboob.X) will fail, etc.
# Instead, we must return the same module to prevent Python from generating
# another one.
# Trick found at https://stackoverflow.com/a/56872393

class AliasLoader(Loader):
    def module_repr(self, module):
        return repr(module)

    def load_module(self, fullname):
        new_name = fullname.replace("weboob", "woob")
        module = importlib.import_module(new_name)
        sys.modules[fullname] = module
        return module


class AliasImporter(MetaPathFinder):
    def find_module(self, fullname, path=None):
        root_name, _, __ = fullname.partition(".")
        if root_name == "weboob":
            return AliasLoader()


sys.meta_path.insert(0, AliasImporter())



# Copyright(C) 2023 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.


from __future__ import annotations

from collections import OrderedDict
from pathlib import Path

from packaging.specifiers import SpecifierSet
from packaging.requirements import Requirement, InvalidRequirement


__all__ = ['parse_requirements']


def parse_requirements(path: str | Path) -> dict[str, SpecifierSet]:
    requirements = OrderedDict()

    try:
        with open(path, 'r', encoding='utf-8') as fp:
            for line in fp.readlines():
                try:
                    r = Requirement(line.strip())
                except InvalidRequirement:
                    # ignore blank lines or comments, but we also don't
                    # want to crash if requirements.txt contains incorrect
                    # lines. So use only this catch-all.
                    continue
                else:
                    requirements[r.name] = r.specifier
    except FileNotFoundError:
        # Assume this works with any version of woob.
        return {}

    return requirements



# Copyright(C) 2010-2013 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from datetime import date as real_date, datetime as real_datetime, timedelta
import time
import re

try:
    import dateutil.parser
    from dateutil import tz
except ImportError:
    raise ImportError('Please install python3-dateutil')


__all__ = [
    'local2utc', 'utc2local',
    'now_as_utc', 'now_as_tz',
    'LinearDateGuesser',
    'date', 'datetime',
    'new_date', 'new_datetime',
    'closest_date',
]


def local2utc(dateobj):
    dateobj = dateobj.replace(tzinfo=tz.tzlocal())
    dateobj = dateobj.astimezone(tz.tzutc())
    return dateobj


def utc2local(dateobj):
    dateobj = dateobj.replace(tzinfo=tz.tzutc())
    dateobj = dateobj.astimezone(tz.tzlocal())
    return dateobj


def now_as_utc():
    return datetime.now(tz.tzutc())


def now_as_tz(tzinfo):
    if isinstance(tzinfo, str):
        tzinfo = tz.gettz(tzinfo)
    return datetime.now(tzinfo)


class date(real_date):
    def strftime(self, fmt):
        return strftime(self, fmt)

    @classmethod
    def from_date(cls, d):
        return cls(d.year, d.month, d.day)


class datetime(real_datetime):
    def strftime(self, fmt):
        return strftime(self, fmt)

    def combine(self, date, time):
        return datetime(date.year, date.month, date.day, time.hour, time.minute, time.microsecond, time.tzinfo)

    def date(self):
        return date(self.year, self.month, self.day)

    @classmethod
    def from_datetime(cls, dt):
        return cls(dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second, dt.microsecond, dt.tzinfo)


def new_date(d):
    """ Generate a safe date from a datetime.date object """
    return date(d.year, d.month, d.day)


def new_datetime(d):
    """
    Generate a safe datetime from a datetime.date or datetime.datetime object
    """
    kw = [d.year, d.month, d.day]
    if isinstance(d, real_datetime):
        kw.extend([d.hour, d.minute, d.second, d.microsecond, d.tzinfo])
    return datetime(*kw)


# No support for strftime's "%s" or "%y".
# Allowed if there's an even number of "%"s because they are escaped.
_illegal_formatting = re.compile(r"((^|[^%])(%%)*%[sy])")


def _findall(text, substr):
    # Also finds overlaps
    sites = []
    i = 0
    while True:
        j = text.find(substr, i)
        if j == -1:
            break
        sites.append(j)
        i = j+1
    return sites


def strftime(dt, fmt):
    if dt.year >= 1900:
        return super(type(dt), dt).strftime(fmt)
    illegal_formatting = _illegal_formatting.search(fmt)
    if illegal_formatting:
        raise TypeError("strftime of dates before 1900 does not handle" + illegal_formatting.group(0))

    year = dt.year
    # For every non-leap year century, advance by
    # 6 years to get into the 28-year repeat cycle
    delta = 2000 - year
    off = 6*(delta // 100 + delta // 400)
    year = year + off

    # Move to around the year 2000
    year = year + ((2000 - year)//28)*28
    timetuple = dt.timetuple()
    s1 = time.strftime(fmt, (year,) + timetuple[1:])
    sites1 = _findall(s1, str(year))

    s2 = time.strftime(fmt, (year+28,) + timetuple[1:])
    sites2 = _findall(s2, str(year+28))

    sites = []
    for site in sites1:
        if site in sites2:
            sites.append(site)

    s = s1
    syear = "%4d" % (dt.year,)
    for site in sites:
        s = s[:site] + syear + s[site+4:]
    return s


def cmp(a, b):
    return (a > b) - (a < b)


class LinearDateGuesser:
    """
    The aim of this class is to guess the exact date object from
    a day and a month, but not a year.

    It works with a start date (default is today), and all dates must be
    sorted from recent to older.
    """

    def __init__(self, current_date=None, date_max_bump=timedelta(31)):
        self.date_max_bump = date_max_bump
        if current_date is None:
            current_date = date.today()
        self.current_date = current_date

    def try_assigning_year(self, day, month, start_year, max_year):
        """
        Tries to create a date object with day, month and start_year and returns
        it.
        If it fails due to the year not matching the day+month combination
        (i.e. due to a ValueError -- TypeError and OverflowError are not
        handled), the previous or next years are tried until max_year is
        reached.
        In case initialization still fails with max_year, this function raises
        a ValueError.
        """
        while True:
            try:
                return date(start_year, month, day)
            except ValueError as e:
                if start_year == max_year:
                    raise e
                start_year += cmp(max_year, start_year)

    def set_current_date(self, current_date):
        self.current_date = current_date

    def guess_date(self, day, month, change_current_date=True):
        """ Returns a date object built from a given day/month pair. """

        today = self.current_date
        # The website only provides dates using the 'DD/MM' string, so we have to
        # determine the most possible year by ourselves. This implies tracking
        # the current date.
        # However, we may also encounter "bumps" in the dates, e.g. "12/11,
        # 10/11, 10/11, 12/11, 09/11", so we have to be, well, quite tolerant,
        # by accepting dates in the near future (say, 7 days) of the current
        # date. (Please, kill me...)
        # We first try to keep the current year
        naively_parsed_date = self.try_assigning_year(day, month, today.year, today.year - 5)
        if (naively_parsed_date.year != today.year):
            # we most likely hit a 29/02 leading to a change of year
            if change_current_date:
                self.set_current_date(naively_parsed_date)
            return naively_parsed_date

        if (naively_parsed_date > today + self.date_max_bump):
            # if the date ends up too far in the future, consider it actually
            # belongs to the previous year
            parsed_date = date(today.year - 1, month, day)
            if change_current_date:
                self.set_current_date(parsed_date)
        elif (naively_parsed_date > today and naively_parsed_date <= today + self.date_max_bump):
            # if the date is in the near future, consider it is a bump
            parsed_date = naively_parsed_date
            # do not keep it as current date though
        else:
            # if the date is in the past, as expected, simply keep it
            parsed_date = naively_parsed_date
            # and make it the new current date
            if change_current_date:
                self.set_current_date(parsed_date)
        return parsed_date


class ChaoticDateGuesser(LinearDateGuesser):
    """
    This class aim to find the guess the date when you know the
    day and month and the minimum year
    """

    def __init__(self, min_date, current_date=None, date_max_bump=timedelta(31)):
        if min_date is None:
            raise ValueError("min_date is not set")
        self.min_date = min_date
        super(ChaoticDateGuesser, self).__init__(current_date, date_max_bump)

    def guess_date(self, day, month):
        """Returns a possible date between min_date and current_date"""
        parsed_date = super(ChaoticDateGuesser, self).guess_date(day, month, False)
        if parsed_date >= self.min_date:
            return parsed_date
        else:
            raise ValueError("%s is inferior to min_date %s" % (parsed_date, self.min_date))


DATE_TRANSLATE_FR = [(re.compile(r'janvier', re.I),             'january'),
                     (re.compile(r'f[eé]vrier', re.I | re.U),   'february'),
                     (re.compile(r'mars', re.I),                'march'),
                     (re.compile(r'avril', re.I),               'april'),
                     (re.compile(r'mai', re.I),                 'may'),
                     (re.compile(r'juin', re.I),                'june'),
                     (re.compile(r'juillet', re.I),             'july'),
                     (re.compile(r'ao[uû]t?', re.I | re.U),     'august'),
                     (re.compile(r'septembre', re.I),           'september'),
                     (re.compile(r'octobre', re.I),             'october'),
                     (re.compile(r'novembre', re.I),            'november'),
                     (re.compile(r'd[eé]cembre', re.I | re.U),  'december'),
                     (re.compile(r'jan\.', re.I),               'january'),
                     (re.compile(r'janv\.', re.I),              'january'),
                     (re.compile(r'\bjan\b', re.I),             'january'),
                     (re.compile(r'f[eé]v\.', re.I | re.U),     'february'),
                     (re.compile(r'f[eé]vr\.', re.I | re.U),    'february'),
                     (re.compile(r'\\bf[eé]v\b', re.I | re.U),  'february'),
                     (re.compile(r'\\bf[eé]vr\b', re.I | re.U), 'february'),
                     (re.compile(r'avr\.', re.I),               'april'),
                     (re.compile(r'\bavr\b', re.I),             'april'),
                     (re.compile(r'juil\.', re.I),              'july'),
                     (re.compile(r'juill\.', re.I),             'july'),
                     (re.compile(r'\bjuil\b', re.I),            'july'),
                     (re.compile(r'sep\.', re.I),               'september'),
                     (re.compile(r'sept\.', re.I),              'september'),
                     (re.compile(r'\bsep\b', re.I),             'september'),
                     (re.compile(r'oct\.', re.I),               'october'),
                     (re.compile(r'\boct\b', re.I),             'october'),
                     (re.compile(r'nov\.', re.I),               'november'),
                     (re.compile(r'\bnov\b', re.I),             'november'),
                     (re.compile(r'd[eé]c\.', re.I | re.U),     'december'),
                     (re.compile(r'\bd[eé]c\b', re.I | re.U),   'december'),
                     (re.compile(r'lundi', re.I),               'monday'),
                     (re.compile(r'mardi', re.I),               'tuesday'),
                     (re.compile(r'mercredi', re.I),            'wednesday'),
                     (re.compile(r'jeudi', re.I),               'thursday'),
                     (re.compile(r'vendredi', re.I),            'friday'),
                     (re.compile(r'samedi', re.I),              'saturday'),
                     (re.compile(r'dimanche', re.I),            'sunday')]


class FrenchParser(dateutil.parser.parserinfo):
    MONTHS = [
        ("janvier", "janv", "jan"),
        ("février", "fevrier", "fevr", "fev", "févr", "fév"),
        ("mars",),
        ("avril", "avr"),
        ("mai",),
        ("juin",),
        ("juillet", "juill", "juil"),
        ("août", "aout", "aou"),
        ("septembre", "sept"),
        ("octobre", "oct"),
        ("novembre", "nov"),
        ("décembre", "decembre", "déc", "dec"),
    ]

    WEEKDAYS = [
        ("lundi",),
        ("mardi",),
        ("mercredi",),
        ("jeudi",),
        ("vendredi",),
        ("samedi",),
        ("dimanche",),
    ]

    JUMP = [
        " ", ".", ",", ";", "-", "/", "'",
        "à", "le", "er",
    ]

    def __init__(self, dayfirst=None, yearfirst=False):
        if dayfirst is None:
            dayfirst = True
        super().__init__(dayfirst, yearfirst)


class ItalianParser(dateutil.parser.parserinfo):
    MONTHS = [
        ("gennaio",),
        ("febbraio",),
        ("marzo",),
        ("aprile",),
        ("maggio",),
        ("giugno",),
        ("luglio",),
        ("agosto", "ago"),
        ("settembre"),
        ("ottobre",),
        ("novembre",),
        ("dicembre",),
    ]

    WEEKDAYS = [
        ("lunedì", "lunedi"),
        ("martedì", "martedi"),
        ("mercoledì", "mercoledi"),
        ("giovedì", "giovedi"),
        ("venerdì", "venerdi"),
        ("sabato",),
        ("domenica",)
    ]


class SpanishParser(dateutil.parser.parserinfo):
    MONTHS = [
        ("enero", "ene"),
        ("febrero", "feb"),
        ("marzo", "mar"),
        ("abril", "abr"),
        ("mayo", "may"),
        ("junio", "junè"),
        ("julio", "jul"),
        ("agosto", "ago"),
        ("septimebre", "sep"),
        ("octubre", "oct"),
        ("noviembre", "nov"),
        ("diciembre", "dec"),
    ]

    WEEKDAYS = [
        ("lunes",),
        ("martes",),
        ("miércoles", "miercoles"),
        ("jueves",),
        ("viernes",),
        ("sábado", "sabado"),
        ("domingo",),
    ]


def parse_french_date(date, **kwargs):
    return dateutil.parser.parse(date, parserinfo=FrenchParser(), **kwargs)


WEEK   = {'MONDAY': 0,
          'TUESDAY': 1,
          'WEDNESDAY': 2,
          'THURSDAY': 3,
          'FRIDAY': 4,
          'SATURDAY': 5,
          'SUNDAY': 6,
          'LUNDI': 0,
          'MARDI': 1,
          'MERCREDI': 2,
          'JEUDI': 3,
          'VENDREDI': 4,
          'SAMEDI': 5,
          'DIMANCHE': 6,
          }


def get_date_from_day(day):
    today = date.today()
    today_day_number = today.weekday()

    requested_day_number = WEEK[day.upper()]

    if today_day_number < requested_day_number:
        day_to_go = requested_day_number - today_day_number
    else:
        day_to_go = 7 - today_day_number + requested_day_number

    requested_date = today + timedelta(day_to_go)
    return date(requested_date.year, requested_date.month, requested_date.day)


def parse_date(string):
    matches = re.search(r'\s*([012]?[0-9]|3[01])\s*/\s*(0?[1-9]|1[012])\s*/?(\d{2}|\d{4})?$', string)
    if matches:
        year = matches.group(3)
        if not year:
            year = date.today().year
        elif len(year) == 2:
            year = 2000 + int(year)
        return date(int(year), int(matches.group(2)), int(matches.group(1)))

    elif string.upper() in list(WEEK.keys()):
        return get_date_from_day(string)

    elif string.upper() == "TODAY":
        return date.today()


def closest_date(date, date_from, date_to):
    """
    Adjusts year so that the date is closest to the given range.
    Transactions dates in a statement usually contain only day and month.
    Statement dates range have a year though.
    Merge them all together to get a full transaction date.
    """
    # If the date is within given range, we're done.
    if date_from <= date <= date_to:
        return date

    dates = [real_datetime(year, date.month, date.day)
             for year in range(date_from.year, date_to.year+1)]

    # Ideally, pick the date within given range.
    for d in dates:
        if date_from <= d <= date_to:
            return d

    # Otherwise, return the most recent date in the past.
    return min(dates, key=lambda d: abs(d-date_from))
